# Portions copyright (c) Meta Platforms, Inc. and affiliates.

# pyre-strict

import sys
from opcode import (
    hascompare,
    hasconst,
    hasfree,
    hasjabs,
    hasjrel,
    haslocal,
    hasname,
    opmap,
    opname,
)

from .opcodebase import Opcode

STATIC_OPMAP: dict[str, int] = {}
STATIC_OPNAMES: list[str] = [f"<{i}>" for i in range(256)]
STATIC_CONST_OPCODES: list[int] = []

if sys.version_info >= (3, 12):
    # pyre-fixme[21]: Could not find name `_cache_format` in `opcode` (stubbed).
    # pyre-fixme[21]: Could not find name `_inline_cache_entries` in `opcode` (stubbed).
    # pyre-fixme[21]: Could not find name `_specializations` in `opcode` (stubbed).
    from opcode import _cache_format, _inline_cache_entries, _specializations, hasarg

    if sys.version_info >= (3, 15):
        opmap = dict(opmap)
        _cache_format = dict(_cache_format)
        _inline_cache_entries = dict(_inline_cache_entries)
        _specializations = dict(_specializations)

    from cinderx import opcode as cinderx_opcode

    cinderx_opcode.init(
        STATIC_OPNAMES if sys.version_info >= (3, 14) else opname,
        STATIC_OPMAP,
        hasname,
        hasjrel,
        hasjabs,
        STATIC_CONST_OPCODES if sys.version_info >= (3, 14) else hasconst,
        hasarg,
        # pyre-fixme[16]: Module `opcode` has no attribute `_cache_format`.
        _cache_format,
        # pyre-fixme[16]: Module `opcode` has no attribute `_specializations`.
        _specializations,
        # pyre-fixme[16]: Module `opcode` has no attribute `_inline_cache_entries`.
        _inline_cache_entries,
    )
    if sys.version_info >= (3, 14):
        opname[126] = "EXTENDED_OPCODE"
        opmap["EXTENDED_OPCODE"] = 126
        hasarg.append(126)

    opmap.update(STATIC_OPMAP)

    if "dis" in sys.modules:
        # Fix up dis module to use the CinderX opcodes
        import dis

        dis._all_opname = list(opname)
        dis._all_opmap = dict(opmap)
        dis._empty_slot = [
            slot
            # pyre-fixme[16]: Module `dis` has no attribute `_all_opname`.
            for slot, name in enumerate(dis._all_opname)
            if name.startswith("<")
        ]


opcode: Opcode = Opcode()
for opname, opnum in opmap.items():
    if opnum in hasname:
        opcode.name_op(opname, opnum)
    elif opnum in hasjrel:
        opcode.jrel_op(opname, opnum)
    elif opnum in hasjabs:
        opcode.jabs_op(opname, opnum)
    else:
        opcode.def_op(opname, opnum)
        if opnum in hasconst:
            opcode.hasconst.add(opname)
        elif opnum in hasfree:
            opcode.hasfree.add(opnum)
        elif opnum in haslocal:
            opcode.haslocal.add(opnum)
        elif opnum in hascompare:
            opcode.hascompare.add(opnum)


for opnum in STATIC_CONST_OPCODES:
    opcode.hasconst.add(STATIC_OPNAMES[opnum])


FVC_MASK = 0x3
FVC_NONE = 0x0
FVC_STR = 0x1
FVC_REPR = 0x2
FVC_ASCII = 0x3
FVS_MASK = 0x4
FVS_HAVE_SPEC = 0x4


opcode.stack_effects.update(
    NOP=0,
    RESUME=0,
    POP_TOP=-1,
    ROT_TWO=0,
    ROT_THREE=0,
    DUP_TOP=1,
    DUP_TOP_TWO=2,
    ROT_FOUR=0,
    UNARY_POSITIVE=0,
    UNARY_NEGATIVE=0,
    UNARY_NOT=0,
    UNARY_INVERT=0,
    BINARY_MATRIX_MULTIPLY=-1,
    INPLACE_MATRIX_MULTIPLY=-1,
    BINARY_POWER=-1,
    BINARY_MULTIPLY=-1,
    BINARY_MODULO=-1,
    BINARY_ADD=-1,
    BINARY_SUBTRACT=-1,
    BINARY_SUBSCR=-1,
    BINARY_FLOOR_DIVIDE=-1,
    BINARY_TRUE_DIVIDE=-1,
    INPLACE_FLOOR_DIVIDE=-1,
    INPLACE_TRUE_DIVIDE=-1,
    GET_LEN=1,
    MATCH_MAPPING=1,
    MATCH_SEQUENCE=1,
    MATCH_KEYS=2,
    COPY_DICT_WITHOUT_KEYS=0,
    WITH_EXCEPT_START=1,
    GET_AITER=0,
    GET_ANEXT=1,
    BEFORE_ASYNC_WITH=1,
    END_ASYNC_FOR=-7,
    INPLACE_ADD=-1,
    INPLACE_SUBTRACT=-1,
    INPLACE_MULTIPLY=-1,
    INPLACE_MODULO=-1,
    STORE_SUBSCR=-3,
    DELETE_SUBSCR=-2,
    BINARY_LSHIFT=-1,
    BINARY_RSHIFT=-1,
    BINARY_AND=-1,
    BINARY_XOR=-1,
    BINARY_OR=-1,
    INPLACE_POWER=-1,
    GET_ITER=0,
    GET_YIELD_FROM_ITER=0,
    PRINT_EXPR=-1,
    LOAD_BUILD_CLASS=1,
    YIELD_FROM=-1,
    GET_AWAITABLE=0,
    LOAD_ASSERTION_ERROR=1,
    INPLACE_LSHIFT=-1,
    INPLACE_RSHIFT=-1,
    INPLACE_AND=-1,
    INPLACE_XOR=-1,
    INPLACE_OR=-1,
    LIST_TO_TUPLE=0,
    RETURN_VALUE=-1,
    IMPORT_STAR=-1,
    SETUP_ANNOTATIONS=0,
    YIELD_VALUE=0,
    POP_BLOCK=0,
    POP_EXCEPT=-3,
    STORE_NAME=-1,
    DELETE_NAME=0,
    UNPACK_SEQUENCE=lambda oparg, jmp=0: oparg - 1,
    FOR_ITER=lambda oparg, jmp=0: -1 if jmp > 0 else 1,
    UNPACK_EX=lambda oparg, jmp=0: (oparg & 0xFF) + (oparg >> 8),
    STORE_ATTR=-2,
    DELETE_ATTR=-1,
    STORE_GLOBAL=-1,
    DELETE_GLOBAL=0,
    ROT_N=0,
    LOAD_CONST=1,
    LOAD_NAME=1,
    BUILD_TUPLE=lambda oparg, jmp=0: 1 - oparg,
    BUILD_LIST=lambda oparg, jmp=0: 1 - oparg,
    BUILD_SET=lambda oparg, jmp=0: 1 - oparg,
    BUILD_MAP=lambda oparg, jmp=0: 1 - 2 * oparg,
    LOAD_ATTR=0,
    COMPARE_OP=-1,
    IMPORT_NAME=-1,
    IMPORT_FROM=1,
    JUMP_FORWARD=0,
    JUMP_IF_FALSE_OR_POP=lambda oparg, jmp=0: 0 if jmp else -1,
    JUMP_IF_TRUE_OR_POP=lambda oparg, jmp=0: 0 if jmp else -1,
    JUMP_ABSOLUTE=0,
    POP_JUMP_IF_FALSE=-1,
    POP_JUMP_IF_TRUE=-1,
    LOAD_GLOBAL=1,
    IS_OP=-1,
    CONTAINS_OP=-1,
    RERAISE=-3,
    JUMP_IF_NOT_EXC_MATCH=-2,
    SETUP_FINALLY=lambda oparg, jmp: 6 if jmp else 0,
    LOAD_FAST=1,
    STORE_FAST=-1,
    DELETE_FAST=0,
    GEN_START=-1,
    RAISE_VARARGS=lambda oparg, jmp=0: -oparg,
    CALL_FUNCTION=lambda oparg, jmp=0: -oparg,
    MAKE_FUNCTION=lambda oparg, jmp=0: -1
    - ((oparg & 0x01) != 0)
    - ((oparg & 0x02) != 0)
    - ((oparg & 0x04) != 0)
    - ((oparg & 0x08) != 0),
    BUILD_SLICE=lambda oparg, jmp=0: -2 if oparg == 3 else -1,
    LOAD_CLOSURE=1,
    LOAD_DEREF=1,
    STORE_DEREF=-1,
    DELETE_DEREF=0,
    CALL_FUNCTION_KW=lambda oparg, jmp=0: -oparg - 1,
    CALL_FUNCTION_EX=lambda oparg, jmp=0: -1 - ((oparg & 0x01) != 0),
    SETUP_WITH=lambda oparg, jmp=0: 6 if jmp else 1,
    EXTENDED_ARG=0,
    LIST_APPEND=-1,
    SET_ADD=-1,
    MAP_ADD=-2,
    LOAD_CLASSDEREF=1,
    MATCH_CLASS=-1,
    SETUP_ASYNC_WITH=lambda oparg, jmp: (-1 + 6) if jmp else 0,
    # If there's a fmt_spec on the stack, we go from 2->1,
    # else 1->1.
    FORMAT_VALUE=lambda oparg, jmp=0: -1 if (oparg & FVS_MASK) == FVS_HAVE_SPEC else 0,
    BUILD_CONST_KEY_MAP=lambda oparg, jmp=0: -oparg,
    BUILD_STRING=lambda oparg, jmp=0: 1 - oparg,
    INVOKE_METHOD=lambda oparg, jmp=0: -(oparg[1] + 1),
    LOAD_METHOD=1,
    CALL_METHOD=lambda oparg, jmp=0: -oparg - 1,
    LIST_EXTEND=-1,
    SET_UPDATE=-1,
    DICT_MERGE=-1,
    DICT_UPDATE=-1,
)

if sys.version_info >= (3, 12):
    # pyre-fixme[21]: Could not find name `_intrinsic_1_descs` in `opcode` (stubbed).
    # pyre-fixme[21]: Could not find name `_intrinsic_2_descs` in `opcode` (stubbed).
    # pyre-fixme[21]: Could not find name `_nb_ops` in `opcode` (stubbed).
    from opcode import (
        _intrinsic_1_descs as INTRINSIC_1,
        _intrinsic_2_descs as INTRINSIC_2,
        _nb_ops as NB_OPS,
    )
else:
    NB_OPS: list[tuple[str, str]] = []
    INTRINSIC_1: list[str] = []
    INTRINSIC_2: list[str] = []


if sys.version_info >= (3, 14):
    opcode.popped.update(
        ANNOTATIONS_PLACEHOLDER=0,
        BINARY_OP=2,
        BINARY_OP_ADD_FLOAT=2,
        BINARY_OP_ADD_INT=2,
        BINARY_OP_ADD_UNICODE=2,
        BINARY_OP_EXTEND=2,
        BINARY_OP_INPLACE_ADD_UNICODE=2,
        BINARY_OP_MULTIPLY_FLOAT=2,
        BINARY_OP_MULTIPLY_INT=2,
        BINARY_OP_SUBSCR_DICT=2,
        BINARY_OP_SUBSCR_GETITEM=2,
        BINARY_OP_SUBSCR_LIST_INT=2,
        BINARY_OP_SUBSCR_LIST_SLICE=2,
        BINARY_OP_SUBSCR_STR_INT=2,
        BINARY_OP_SUBSCR_TUPLE_INT=2,
        BINARY_OP_SUBTRACT_FLOAT=2,
        BINARY_OP_SUBTRACT_INT=2,
        BINARY_SLICE=3,
        BUILD_INTERPOLATION=lambda oparg: 2 + (oparg & 1),
        BUILD_LIST=lambda oparg: oparg,
        BUILD_MAP=lambda oparg: oparg * 2,
        BUILD_SET=lambda oparg: oparg,
        BUILD_SLICE=lambda oparg: oparg,
        BUILD_STRING=lambda oparg: oparg,
        BUILD_TEMPLATE=2,
        BUILD_TUPLE=lambda oparg: oparg,
        CACHE=0,
        CALL=lambda oparg: 2 + oparg,
        CALL_ALLOC_AND_ENTER_INIT=lambda oparg: 2 + oparg,
        CALL_BOUND_METHOD_EXACT_ARGS=lambda oparg: 2 + oparg,
        CALL_BOUND_METHOD_GENERAL=lambda oparg: 2 + oparg,
        CALL_BUILTIN_CLASS=lambda oparg: 2 + oparg,
        CALL_BUILTIN_FAST=lambda oparg: 2 + oparg,
        CALL_BUILTIN_FAST_WITH_KEYWORDS=lambda oparg: 2 + oparg,
        CALL_BUILTIN_O=lambda oparg: 2 + oparg,
        CALL_FUNCTION_EX=4,
        CALL_INTRINSIC_1=1,
        CALL_INTRINSIC_2=2,
        CALL_ISINSTANCE=lambda oparg: 2 + oparg,
        CALL_KW=lambda oparg: 3 + oparg,
        CALL_KW_BOUND_METHOD=lambda oparg: 3 + oparg,
        CALL_KW_NON_PY=lambda oparg: 3 + oparg,
        CALL_KW_PY=lambda oparg: 3 + oparg,
        CALL_LEN=3,
        CALL_LIST_APPEND=3,
        CALL_METHOD_DESCRIPTOR_FAST=lambda oparg: 2 + oparg,
        CALL_METHOD_DESCRIPTOR_FAST_WITH_KEYWORDS=lambda oparg: 2 + oparg,
        CALL_METHOD_DESCRIPTOR_NOARGS=lambda oparg: 2 + oparg,
        CALL_METHOD_DESCRIPTOR_O=lambda oparg: 2 + oparg,
        CALL_NON_PY_GENERAL=lambda oparg: 2 + oparg,
        CALL_PY_EXACT_ARGS=lambda oparg: 2 + oparg,
        CALL_PY_GENERAL=lambda oparg: 2 + oparg,
        CALL_STR_1=3,
        CALL_TUPLE_1=3,
        CALL_TYPE_1=3,
        CHECK_EG_MATCH=2,
        CHECK_EXC_MATCH=2,
        CLEANUP_THROW=3,
        COMPARE_OP=2,
        COMPARE_OP_FLOAT=2,
        COMPARE_OP_INT=2,
        COMPARE_OP_STR=2,
        CONTAINS_OP=2,
        CONTAINS_OP_DICT=2,
        CONTAINS_OP_SET=2,
        CONVERT_VALUE=1,
        COPY=lambda oparg: 1 + (oparg - 1),
        COPY_FREE_VARS=0,
        DELETE_ATTR=1,
        DELETE_DEREF=0,
        DELETE_FAST=0,
        DELETE_GLOBAL=0,
        DELETE_NAME=0,
        DELETE_SUBSCR=2,
        DICT_MERGE=lambda oparg: 5 + (oparg - 1),
        DICT_UPDATE=lambda oparg: 2 + (oparg - 1),
        EAGER_IMPORT_NAME=2,
        END_ASYNC_FOR=2,
        END_FOR=1,
        END_SEND=2,
        ENTER_EXECUTOR=0,
        EXIT_INIT_CHECK=1,
        EXTENDED_ARG=0,
        FORMAT_SIMPLE=1,
        FORMAT_WITH_SPEC=2,
        FOR_ITER=1,
        FOR_ITER_GEN=1,
        FOR_ITER_LIST=1,
        FOR_ITER_RANGE=1,
        FOR_ITER_TUPLE=1,
        GET_AITER=1,
        GET_ANEXT=1,
        GET_AWAITABLE=1,
        GET_ITER=1,
        GET_LEN=1,
        GET_YIELD_FROM_ITER=1,
        IMPORT_FROM=1,
        IMPORT_NAME=2,
        INSTRUMENTED_CALL=lambda oparg: 2 + oparg,
        INSTRUMENTED_CALL_FUNCTION_EX=4,
        INSTRUMENTED_CALL_KW=lambda oparg: 3 + oparg,
        INSTRUMENTED_END_ASYNC_FOR=2,
        INSTRUMENTED_END_FOR=2,
        INSTRUMENTED_END_SEND=2,
        INSTRUMENTED_FOR_ITER=1,
        INSTRUMENTED_INSTRUCTION=0,
        INSTRUMENTED_JUMP_BACKWARD=0,
        INSTRUMENTED_JUMP_FORWARD=0,
        INSTRUMENTED_LINE=0,
        INSTRUMENTED_LOAD_SUPER_ATTR=3,
        INSTRUMENTED_NOT_TAKEN=0,
        INSTRUMENTED_POP_ITER=1,
        INSTRUMENTED_POP_JUMP_IF_FALSE=1,
        INSTRUMENTED_POP_JUMP_IF_NONE=1,
        INSTRUMENTED_POP_JUMP_IF_NOT_NONE=1,
        INSTRUMENTED_POP_JUMP_IF_TRUE=1,
        INSTRUMENTED_RESUME=0,
        INSTRUMENTED_RETURN_VALUE=1,
        INSTRUMENTED_YIELD_VALUE=1,
        INTERPRETER_EXIT=1,
        IS_OP=2,
        JUMP=0,
        JUMP_BACKWARD=0,
        JUMP_BACKWARD_JIT=0,
        JUMP_BACKWARD_NO_INTERRUPT=0,
        JUMP_BACKWARD_NO_JIT=0,
        JUMP_FORWARD=0,
        JUMP_IF_FALSE=1,
        JUMP_IF_TRUE=1,
        JUMP_NO_INTERRUPT=0,
        LIST_APPEND=lambda oparg: 2 + (oparg - 1),
        LIST_EXTEND=lambda oparg: 2 + (oparg - 1),
        LOAD_ATTR=1,
        LOAD_ATTR_CLASS=1,
        LOAD_ATTR_CLASS_WITH_METACLASS_CHECK=1,
        LOAD_ATTR_GETATTRIBUTE_OVERRIDDEN=1,
        LOAD_ATTR_INSTANCE_VALUE=1,
        LOAD_ATTR_METHOD_LAZY_DICT=1,
        LOAD_ATTR_METHOD_NO_DICT=1,
        LOAD_ATTR_METHOD_WITH_VALUES=1,
        LOAD_ATTR_MODULE=1,
        LOAD_ATTR_NONDESCRIPTOR_NO_DICT=1,
        LOAD_ATTR_NONDESCRIPTOR_WITH_VALUES=1,
        LOAD_ATTR_PROPERTY=1,
        LOAD_ATTR_SLOT=1,
        LOAD_ATTR_WITH_HINT=1,
        LOAD_BUILD_CLASS=0,
        LOAD_CLOSURE=0,
        LOAD_COMMON_CONSTANT=0,
        LOAD_CONST=0,
        LOAD_CONST_IMMORTAL=0,
        LOAD_CONST_MORTAL=0,
        LOAD_DEREF=0,
        LOAD_FAST=0,
        LOAD_FAST_AND_CLEAR=0,
        LOAD_FAST_BORROW=0,
        LOAD_FAST_BORROW_LOAD_FAST_BORROW=0,
        LOAD_FAST_CHECK=0,
        LOAD_FAST_LOAD_FAST=0,
        LOAD_FROM_DICT_OR_DEREF=1,
        LOAD_FROM_DICT_OR_GLOBALS=1,
        LOAD_GLOBAL=0,
        LOAD_GLOBAL_BUILTIN=0,
        LOAD_GLOBAL_MODULE=0,
        LOAD_LOCALS=0,
        LOAD_NAME=0,
        LOAD_SMALL_INT=0,
        LOAD_SPECIAL=1,
        LOAD_SUPER_ATTR=3,
        LOAD_SUPER_ATTR_ATTR=3,
        LOAD_SUPER_ATTR_METHOD=3,
        MAKE_CELL=0,
        MAKE_FUNCTION=1,
        MAP_ADD=lambda oparg: 3 + (oparg - 1),
        MATCH_CLASS=3,
        MATCH_KEYS=2,
        MATCH_MAPPING=1,
        MATCH_SEQUENCE=1,
        NOP=0,
        NOT_TAKEN=0,
        POP_BLOCK=0,
        POP_EXCEPT=1,
        POP_ITER=1,
        POP_JUMP_IF_FALSE=1,
        POP_JUMP_IF_NONE=1,
        POP_JUMP_IF_NOT_NONE=1,
        POP_JUMP_IF_TRUE=1,
        POP_TOP=1,
        PUSH_EXC_INFO=1,
        PUSH_NULL=0,
        RAISE_VARARGS=lambda oparg: oparg,
        RERAISE=lambda oparg: 1 + oparg,
        RESERVED=0,
        RESUME=0,
        RESUME_CHECK=0,
        RETURN_GENERATOR=0,
        RETURN_VALUE=1,
        SEND=2,
        SEND_GEN=2,
        SETUP_ANNOTATIONS=0,
        SETUP_CLEANUP=0,
        SETUP_FINALLY=0,
        SETUP_WITH=0,
        SET_ADD=lambda oparg: 2 + (oparg - 1),
        SET_FUNCTION_ATTRIBUTE=2,
        SET_UPDATE=lambda oparg: 2 + (oparg - 1),
        STORE_ATTR=2,
        STORE_ATTR_INSTANCE_VALUE=2,
        STORE_ATTR_SLOT=2,
        STORE_ATTR_WITH_HINT=2,
        STORE_DEREF=1,
        STORE_FAST=1,
        STORE_FAST_LOAD_FAST=1,
        STORE_FAST_MAYBE_NULL=1,
        STORE_FAST_STORE_FAST=2,
        STORE_GLOBAL=1,
        STORE_NAME=1,
        STORE_SLICE=4,
        STORE_SUBSCR=3,
        STORE_SUBSCR_DICT=3,
        STORE_SUBSCR_LIST_INT=3,
        SWAP=lambda oparg: 2 + (oparg - 2),
        TO_BOOL=1,
        TO_BOOL_ALWAYS_TRUE=1,
        TO_BOOL_BOOL=1,
        TO_BOOL_INT=1,
        TO_BOOL_LIST=1,
        TO_BOOL_NONE=1,
        TO_BOOL_STR=1,
        UNARY_INVERT=1,
        UNARY_NEGATIVE=1,
        UNARY_NOT=1,
        UNPACK_EX=1,
        UNPACK_SEQUENCE=1,
        UNPACK_SEQUENCE_LIST=1,
        UNPACK_SEQUENCE_TUPLE=1,
        UNPACK_SEQUENCE_TWO_TUPLE=1,
        WITH_EXCEPT_START=5,
        YIELD_VALUE=1,
        # Static opcodes
        EXTENDED_OPCODE=0,
        LOAD_FIELD=1,
        STORE_FIELD=2,
        CAST=1,
        LOAD_LOCAL=0,
        STORE_LOCAL=1,
        PRIMITIVE_BOX=1,
        POP_JUMP_IF_ZERO=1,
        POP_JUMP_IF_NONZERO=1,
        PRIMITIVE_UNBOX=1,
        PRIMITIVE_BINARY_OP=2,
        PRIMITIVE_UNARY_OP=1,
        PRIMITIVE_COMPARE_OP=2,
        LOAD_ITERABLE_ARG=1,
        LOAD_MAPPING_ARG=lambda oparg, jmp=0: 3 if oparg == 3 else 2,
        INVOKE_METHOD=lambda oparg, jmp=0: (oparg[1] + 2),
        INVOKE_FUNCTION=lambda oparg, jmp=0: (oparg[1]),
        INVOKE_NATIVE=lambda oparg, jmp=0: (len(oparg[1]) - 1),
        JUMP_IF_ZERO_OR_POP=lambda oparg, jmp=0: 0 if jmp else -1,
        JUMP_IF_NONZERO_OR_POP=lambda oparg, jmp=0: 0 if jmp else -1,
        FAST_LEN=1,
        CONVERT_PRIMITIVE=1,
        LOAD_CLASS=0,
        BUILD_CHECKED_MAP=lambda oparg, jmp=0: 2 * oparg[1],
        SEQUENCE_GET=2,
        SEQUENCE_SET=3,
        LIST_DEL=2,
        REFINE_TYPE=1,
        PRIMITIVE_LOAD_CONST=0,
        RETURN_PRIMITIVE=1,
        TP_ALLOC=0,
        BUILD_CHECKED_LIST=lambda oparg, jmp=0: oparg[1],
        LOAD_TYPE=1,
        LOAD_METHOD_STATIC=1,
    )

    opcode.pushed.update(
        ANNOTATIONS_PLACEHOLDER=0,
        BINARY_OP=1,
        BINARY_OP_ADD_FLOAT=1,
        BINARY_OP_ADD_INT=1,
        BINARY_OP_ADD_UNICODE=1,
        BINARY_OP_EXTEND=1,
        BINARY_OP_INPLACE_ADD_UNICODE=0,
        BINARY_OP_MULTIPLY_FLOAT=1,
        BINARY_OP_MULTIPLY_INT=1,
        BINARY_OP_SUBSCR_DICT=1,
        BINARY_OP_SUBSCR_GETITEM=0,
        BINARY_OP_SUBSCR_LIST_INT=1,
        BINARY_OP_SUBSCR_LIST_SLICE=1,
        BINARY_OP_SUBSCR_STR_INT=1,
        BINARY_OP_SUBSCR_TUPLE_INT=1,
        BINARY_OP_SUBTRACT_FLOAT=1,
        BINARY_OP_SUBTRACT_INT=1,
        BINARY_SLICE=1,
        BUILD_INTERPOLATION=1,
        BUILD_LIST=1,
        BUILD_MAP=1,
        BUILD_SET=1,
        BUILD_SLICE=1,
        BUILD_STRING=1,
        BUILD_TEMPLATE=1,
        BUILD_TUPLE=1,
        CACHE=0,
        CALL=1,
        CALL_ALLOC_AND_ENTER_INIT=0,
        CALL_BOUND_METHOD_EXACT_ARGS=0,
        CALL_BOUND_METHOD_GENERAL=0,
        CALL_BUILTIN_CLASS=1,
        CALL_BUILTIN_FAST=1,
        CALL_BUILTIN_FAST_WITH_KEYWORDS=1,
        CALL_BUILTIN_O=1,
        CALL_FUNCTION_EX=1,
        CALL_INTRINSIC_1=1,
        CALL_INTRINSIC_2=1,
        CALL_ISINSTANCE=1,
        CALL_KW=1,
        CALL_KW_BOUND_METHOD=0,
        CALL_KW_NON_PY=1,
        CALL_KW_PY=0,
        CALL_LEN=1,
        CALL_LIST_APPEND=0,
        CALL_METHOD_DESCRIPTOR_FAST=1,
        CALL_METHOD_DESCRIPTOR_FAST_WITH_KEYWORDS=1,
        CALL_METHOD_DESCRIPTOR_NOARGS=1,
        CALL_METHOD_DESCRIPTOR_O=1,
        CALL_NON_PY_GENERAL=1,
        CALL_PY_EXACT_ARGS=0,
        CALL_PY_GENERAL=0,
        CALL_STR_1=1,
        CALL_TUPLE_1=1,
        CALL_TYPE_1=1,
        CHECK_EG_MATCH=2,
        CHECK_EXC_MATCH=2,
        CLEANUP_THROW=2,
        COMPARE_OP=1,
        COMPARE_OP_FLOAT=1,
        COMPARE_OP_INT=1,
        COMPARE_OP_STR=1,
        CONTAINS_OP=1,
        CONTAINS_OP_DICT=1,
        CONTAINS_OP_SET=1,
        CONVERT_VALUE=1,
        COPY=lambda oparg: 2 + (oparg - 1),
        COPY_FREE_VARS=0,
        DELETE_ATTR=0,
        DELETE_DEREF=0,
        DELETE_FAST=0,
        DELETE_GLOBAL=0,
        DELETE_NAME=0,
        DELETE_SUBSCR=0,
        DICT_MERGE=lambda oparg: 4 + (oparg - 1),
        DICT_UPDATE=lambda oparg: 1 + (oparg - 1),
        EAGER_IMPORT_NAME=1,
        END_ASYNC_FOR=0,
        END_FOR=0,
        END_SEND=1,
        ENTER_EXECUTOR=0,
        EXIT_INIT_CHECK=0,
        EXTENDED_ARG=0,
        FORMAT_SIMPLE=1,
        FORMAT_WITH_SPEC=1,
        FOR_ITER=2,
        FOR_ITER_GEN=1,
        FOR_ITER_LIST=2,
        FOR_ITER_RANGE=2,
        FOR_ITER_TUPLE=2,
        GET_AITER=1,
        GET_ANEXT=2,
        GET_AWAITABLE=1,
        GET_ITER=1,
        GET_LEN=2,
        GET_YIELD_FROM_ITER=1,
        IMPORT_FROM=2,
        IMPORT_NAME=1,
        INSTRUMENTED_CALL=1,
        INSTRUMENTED_CALL_FUNCTION_EX=1,
        INSTRUMENTED_CALL_KW=1,
        INSTRUMENTED_END_ASYNC_FOR=0,
        INSTRUMENTED_END_FOR=1,
        INSTRUMENTED_END_SEND=1,
        INSTRUMENTED_FOR_ITER=2,
        INSTRUMENTED_INSTRUCTION=0,
        INSTRUMENTED_JUMP_BACKWARD=0,
        INSTRUMENTED_JUMP_FORWARD=0,
        INSTRUMENTED_LINE=0,
        INSTRUMENTED_LOAD_SUPER_ATTR=lambda oparg: 1 + (oparg & 1),
        INSTRUMENTED_NOT_TAKEN=0,
        INSTRUMENTED_POP_ITER=0,
        INSTRUMENTED_POP_JUMP_IF_FALSE=0,
        INSTRUMENTED_POP_JUMP_IF_NONE=0,
        INSTRUMENTED_POP_JUMP_IF_NOT_NONE=0,
        INSTRUMENTED_POP_JUMP_IF_TRUE=0,
        INSTRUMENTED_RESUME=0,
        INSTRUMENTED_RETURN_VALUE=1,
        INSTRUMENTED_YIELD_VALUE=1,
        INTERPRETER_EXIT=0,
        IS_OP=1,
        JUMP=0,
        JUMP_BACKWARD=0,
        JUMP_BACKWARD_JIT=0,
        JUMP_BACKWARD_NO_INTERRUPT=0,
        JUMP_BACKWARD_NO_JIT=0,
        JUMP_FORWARD=0,
        JUMP_IF_FALSE=1,
        JUMP_IF_TRUE=1,
        JUMP_NO_INTERRUPT=0,
        LIST_APPEND=lambda oparg: 1 + (oparg - 1),
        LIST_EXTEND=lambda oparg: 1 + (oparg - 1),
        LOAD_ATTR=lambda oparg: (2 if (isinstance(oparg, tuple) and oparg[1]) else 1),
        LOAD_ATTR_CLASS=lambda oparg: 1 + (oparg & 1),
        LOAD_ATTR_CLASS_WITH_METACLASS_CHECK=lambda oparg: 1 + (oparg & 1),
        LOAD_ATTR_GETATTRIBUTE_OVERRIDDEN=1,
        LOAD_ATTR_INSTANCE_VALUE=lambda oparg: 1 + (oparg & 1),
        LOAD_ATTR_METHOD_LAZY_DICT=2,
        LOAD_ATTR_METHOD_NO_DICT=2,
        LOAD_ATTR_METHOD_WITH_VALUES=2,
        LOAD_ATTR_MODULE=lambda oparg: 1 + (oparg & 1),
        LOAD_ATTR_NONDESCRIPTOR_NO_DICT=1,
        LOAD_ATTR_NONDESCRIPTOR_WITH_VALUES=1,
        LOAD_ATTR_PROPERTY=0,
        LOAD_ATTR_SLOT=lambda oparg: 1 + (oparg & 1),
        LOAD_ATTR_WITH_HINT=lambda oparg: 1 + (oparg & 1),
        LOAD_BUILD_CLASS=1,
        LOAD_CLOSURE=1,
        LOAD_COMMON_CONSTANT=1,
        LOAD_CONST=1,
        LOAD_CONST_IMMORTAL=1,
        LOAD_CONST_MORTAL=1,
        LOAD_DEREF=1,
        LOAD_FAST=1,
        LOAD_FAST_AND_CLEAR=1,
        LOAD_FAST_BORROW=1,
        LOAD_FAST_BORROW_LOAD_FAST_BORROW=2,
        LOAD_FAST_CHECK=1,
        LOAD_FAST_LOAD_FAST=2,
        LOAD_FROM_DICT_OR_DEREF=1,
        LOAD_FROM_DICT_OR_GLOBALS=1,
        LOAD_GLOBAL=lambda oparg: 2 if isinstance(oparg, tuple) else 1,
        LOAD_GLOBAL_BUILTIN=lambda oparg: 1 + (oparg & 1),
        LOAD_GLOBAL_MODULE=lambda oparg: 1 + (oparg & 1),
        LOAD_LOCALS=1,
        LOAD_NAME=1,
        LOAD_SMALL_INT=1,
        LOAD_SPECIAL=2,
        LOAD_SUPER_ATTR=lambda oparg: (
            2 if oparg[0] in ("LOAD_SUPER_METHOD", "LOAD_ZERO_SUPER_METHOD") else 1
        ),
        LOAD_SUPER_ATTR_ATTR=1,
        LOAD_SUPER_ATTR_METHOD=2,
        MAKE_CELL=0,
        MAKE_FUNCTION=1,
        MAP_ADD=lambda oparg: 1 + (oparg - 1),
        MATCH_CLASS=1,
        MATCH_KEYS=3,
        MATCH_MAPPING=2,
        MATCH_SEQUENCE=2,
        NOP=0,
        NOT_TAKEN=0,
        POP_BLOCK=0,
        POP_EXCEPT=0,
        POP_ITER=0,
        POP_JUMP_IF_FALSE=0,
        POP_JUMP_IF_NONE=0,
        POP_JUMP_IF_NOT_NONE=0,
        POP_JUMP_IF_TRUE=0,
        POP_TOP=0,
        PUSH_EXC_INFO=2,
        PUSH_NULL=1,
        RAISE_VARARGS=0,
        RERAISE=lambda oparg: oparg,
        RESERVED=0,
        RESUME=0,
        RESUME_CHECK=0,
        RETURN_GENERATOR=1,
        RETURN_VALUE=1,
        SEND=2,
        SEND_GEN=1,
        SETUP_ANNOTATIONS=0,
        SETUP_CLEANUP=2,
        SETUP_FINALLY=1,
        SETUP_WITH=1,
        SET_ADD=lambda oparg: 1 + (oparg - 1),
        SET_FUNCTION_ATTRIBUTE=1,
        SET_UPDATE=lambda oparg: 1 + (oparg - 1),
        STORE_ATTR=0,
        STORE_ATTR_INSTANCE_VALUE=0,
        STORE_ATTR_SLOT=0,
        STORE_ATTR_WITH_HINT=0,
        STORE_DEREF=0,
        STORE_FAST=0,
        STORE_FAST_LOAD_FAST=1,
        STORE_FAST_MAYBE_NULL=0,
        STORE_FAST_STORE_FAST=0,
        STORE_GLOBAL=0,
        STORE_NAME=0,
        STORE_SLICE=0,
        STORE_SUBSCR=0,
        STORE_SUBSCR_DICT=0,
        STORE_SUBSCR_LIST_INT=0,
        SWAP=lambda oparg: 2 + (oparg - 2),
        TO_BOOL=1,
        TO_BOOL_ALWAYS_TRUE=1,
        TO_BOOL_BOOL=1,
        TO_BOOL_INT=1,
        TO_BOOL_LIST=1,
        TO_BOOL_NONE=1,
        TO_BOOL_STR=1,
        UNARY_INVERT=1,
        UNARY_NEGATIVE=1,
        UNARY_NOT=1,
        UNPACK_EX=lambda oparg: 1 + (oparg & 0xFF) + (oparg >> 8),
        UNPACK_SEQUENCE=lambda oparg: oparg,
        UNPACK_SEQUENCE_LIST=lambda oparg: oparg,
        UNPACK_SEQUENCE_TUPLE=lambda oparg: oparg,
        UNPACK_SEQUENCE_TWO_TUPLE=2,
        WITH_EXCEPT_START=6,
        YIELD_VALUE=1,
        # Static opcodes
        EXTENDED_OPCODE=0,
        LOAD_FIELD=1,
        STORE_FIELD=0,
        CAST=1,
        LOAD_LOCAL=1,
        STORE_LOCAL=0,
        PRIMITIVE_BOX=1,
        POP_JUMP_IF_ZERO=0,
        POP_JUMP_IF_NONZERO=0,
        PRIMITIVE_UNBOX=1,
        PRIMITIVE_BINARY_OP=1,
        PRIMITIVE_UNARY_OP=1,
        PRIMITIVE_COMPARE_OP=1,
        LOAD_ITERABLE_ARG=2,
        LOAD_MAPPING_ARG=1,
        INVOKE_METHOD=1,
        INVOKE_FUNCTION=1,
        INVOKE_NATIVE=1,
        JUMP_IF_ZERO_OR_POP=0,
        JUMP_IF_NONZERO_OR_POP=0,
        FAST_LEN=1,
        CONVERT_PRIMITIVE=1,
        LOAD_CLASS=1,
        BUILD_CHECKED_MAP=1,
        SEQUENCE_GET=1,
        SEQUENCE_SET=0,
        LIST_DEL=0,
        REFINE_TYPE=1,
        PRIMITIVE_LOAD_CONST=1,
        RETURN_PRIMITIVE=0,
        TP_ALLOC=1,
        BUILD_CHECKED_LIST=1,
        LOAD_TYPE=1,
        LOAD_METHOD_STATIC=2,
    )

    if sys.version_info >= (3, 15):
        opcode.popped.update(
            FOR_ITER=2,
            POP_ITER=2,
        )
        opcode.pushed.update(
            FOR_ITER=3,
            GET_ITER=2,
        )
    for opname, popped in opcode.popped.items():
        pushed = opcode.pushed[opname]
        if isinstance(popped, int) and isinstance(pushed, int):
            opcode.stack_effects[opname] = pushed - popped
        elif isinstance(popped, int):

            def make_effect(pushed, popped):
                return lambda oparg, jmp=0: pushed(oparg) - popped

            opcode.stack_effects[opname] = make_effect(pushed, popped)
        elif isinstance(pushed, int):

            def make_effect(pushed, popped):
                return lambda oparg, jmp=0: pushed - popped(oparg)

            opcode.stack_effects[opname] = make_effect(pushed, popped)
        else:

            def make_effect(pushed, popped):
                return lambda oparg, jmp=0: pushed(oparg) - popped(oparg)

            opcode.stack_effects[opname] = make_effect(pushed, popped)

elif sys.version_info >= (3, 12):

    def _load_mapping_arg_effect(oparg: int, _jmp: int = 0) -> int:
        """The effect of this opcode depends on the number of items on the stack
        when it is invoked. The number of items on the stack for LOAD_MAPPING_ARG
        is based on the oparg. LOAD_MAPPING_ARG will push the resulting value and
        will consume either a default value, the mapping, and the argument name when
        oparg == 3 or just the mapping and argument name when there is no default
        value for the argument."""
        if oparg == 2:
            return -1
        elif oparg == 3:
            return -2
        raise ValueError("bad oparg")

    opcode.stack_effects.update(
        CALL=lambda oparg, jmp=0: -(oparg + 1),
        COPY_FREE_VARS=lambda oparg, jmp=0: 0,
        MAKE_CELL=lambda oparg, jmp=0: 0,
        JUMP=0,
        JUMP_BACKWARD=0,
        PUSH_NULL=1,
        RETURN_CONST=0,
        SWAP=0,
        COPY=1,
        CALL_INTRINSIC_1=0,
        CALL_INTRINSIC_2=-1,
        LOAD_FROM_DICT_OR_DEREF=0,
        LOAD_LOCALS=1,
        MAKE_FUNCTION=lambda oparg, jmp=0: -oparg.bit_count(),
        LOAD_FROM_DICT_OR_GLOBALS=0,
        LOAD_SUPER_ATTR=lambda oparg, jmp: (
            2 if oparg[0] in ("LOAD_SUPER_METHOD", "LOAD_ZERO_SUPER_METHOD") else 1
        )
        - 3,
        LOAD_ATTR=lambda oparg, jmp: (
            1 if (isinstance(oparg, tuple) and oparg[1]) else 0
        ),
        BINARY_OP=-1,
        RETURN_GENERATOR=0,
        SEND=0,
        JUMP_NO_INTERRUPT=0,
        JUMP_BACKWARD_NO_INTERRUPT=0,
        CLEANUP_THROW=-1,
        END_SEND=-1,
        END_FOR=-2,
        FOR_ITER=lambda oparg, jmp=0: 1,
        END_ASYNC_FOR=-2,
        SETUP_FINALLY=lambda oparg, jmp=0: 1 if jmp else 0,
        SETUP_CLEANUP=lambda oparg, jmp=0: 2 if jmp else 0,
        SETUP_WITH=lambda oparg, jmp=0: 1 if jmp else 0,
        PUSH_EXC_INFO=1,
        POP_EXCEPT=-1,
        LOAD_FAST_AND_CLEAR=1,
        STORE_FAST_MAYBE_NULL=-1,
        RERAISE=-1,
        LOAD_FAST_CHECK=1,
        LOAD_GLOBAL=lambda oparg, jmp=0: 2 if isinstance(oparg, tuple) else 1,
        BEFORE_WITH=1,
        CALL_FUNCTION_EX=lambda oparg, jmp=0: 1 - (4 if oparg & 0x01 else 3),
        CHECK_EXC_MATCH=0,
        CHECK_EG_MATCH=0,
        POP_JUMP_IF_NONE=-1,
        POP_JUMP_IF_NOT_NONE=-1,
        BINARY_SLICE=-2,
        STORE_SLICE=-4,
        KW_NAMES=0,
        MATCH_KEYS=1,
        MATCH_CLASS=-2,
        EAGER_IMPORT_NAME=-1,
        LOAD_FIELD=0,
        STORE_FIELD=-2,
        CAST=0,
        LOAD_LOCAL=1,
        STORE_LOCAL=-1,
        PRIMITIVE_BOX=0,
        POP_JUMP_IF_ZERO=-1,
        POP_JUMP_IF_NONZERO=-1,
        PRIMITIVE_UNBOX=0,
        PRIMITIVE_BINARY_OP=lambda oparg, jmp: -1,
        PRIMITIVE_UNARY_OP=lambda oparg, jmp: 0,
        PRIMITIVE_COMPARE_OP=lambda oparg, jmp: -1,
        LOAD_ITERABLE_ARG=1,
        LOAD_MAPPING_ARG=_load_mapping_arg_effect,
        INVOKE_FUNCTION=lambda oparg, jmp=0: (-oparg[1]) + 1,
        INVOKE_NATIVE=lambda oparg, jmp=0: (-len(oparg[1])) + 2,
        JUMP_IF_ZERO_OR_POP=lambda oparg, jmp=0: 0 if jmp else -1,
        JUMP_IF_NONZERO_OR_POP=lambda oparg, jmp=0: 0 if jmp else -1,
        FAST_LEN=0,
        CONVERT_PRIMITIVE=0,
        LOAD_CLASS=1,
        BUILD_CHECKED_MAP=lambda oparg, jmp: 1 - 2 * oparg[1],
        SEQUENCE_GET=-1,
        SEQUENCE_SET=-3,
        LIST_DEL=-2,
        REFINE_TYPE=0,
        PRIMITIVE_LOAD_CONST=1,
        RETURN_PRIMITIVE=-1,
        TP_ALLOC=1,
        BUILD_CHECKED_LIST=lambda oparg, jmp: 1 - oparg[1],
        LOAD_TYPE=0,
        LOAD_METHOD_STATIC=1,
    )


def find_op_idx(opname: str) -> int:
    # pyre-fixme[16]: Module `opcode` has no attribute `_nb_ops`.
    for i, (name, _symbol) in enumerate(NB_OPS):
        if name == opname:
            return i

    return -1


STATIC_OPCODES = {
    "LOAD_FIELD",
    "STORE_FIELD",
    "CAST",
    "LOAD_LOCAL",
    "STORE_LOCAL",
    "PRIMITIVE_BOX",
    "POP_JUMP_IF_ZERO",
    "POP_JUMP_IF_NONZERO",
    "PRIMITIVE_UNBOX",
    "PRIMITIVE_BINARY_OP",
    "PRIMITIVE_UNARY_OP",
    "PRIMITIVE_COMPARE_OP",
    "LOAD_ITERABLE_ARG",
    "LOAD_MAPPING_ARG",
    "INVOKE_FUNCTION",
    "INVOKE_METHOD",
    "INVOKE_NATIVE",
    "JUMP_IF_ZERO_OR_POP",
    "JUMP_IF_NONZERO_OR_POP",
    "FAST_LEN",
    "CONVERT_PRIMITIVE",
    "LOAD_CLASS",
    "BUILD_CHECKED_MAP",
    "SEQUENCE_GET",
    "SEQUENCE_SET",
    "LIST_DEL",
    "REFINE_TYPE",
    "PRIMITIVE_LOAD_CONST",
    "RETURN_PRIMITIVE",
    "TP_ALLOC",
    "BUILD_CHECKED_LIST",
    "LOAD_TYPE",
    "LOAD_METHOD_STATIC",
}



# Copyright (c) Meta Platforms, Inc. and affiliates.

# pyre-strict
"""Debugging output for various internal datatypes."""

from __future__ import annotations

from typing import TYPE_CHECKING

from .opcodes import opcode

if TYPE_CHECKING:
    from .pyassem import Block, Instruction, PyFlowGraph


def str_of_oparg(instr: Instruction) -> str:
    if instr.target is None:
        # This is not a block
        return str(instr.oparg)
    elif instr.target.label:
        # This is a labelled block
        return f"{instr.target.bid} ({instr.target.label})"
    else:
        # This is an unlabelled block
        return str(instr.target.bid)


def str_of_instr(instr: Instruction) -> str:
    oparg = str_of_oparg(instr)
    opname = instr.opname
    return f"{instr.lineno} {opname} {oparg}"


def str_of_block_header(block: Block) -> str:
    return repr(block)


def str_of_stack_effect(instr: Instruction) -> str:
    d1 = opcode.stack_effect_raw(instr.opname, instr.oparg, False)
    d2 = opcode.stack_effect_raw(instr.opname, instr.oparg, True)
    if d1 != d2:
        delta = f"{d1}/{d2}"
    else:
        delta = str(d1)
    return delta


def str_of_block_instr(
    instr: Instruction, pc: int = 0, stack_effect: bool = False
) -> str:
    delta = str_of_stack_effect(instr) if stack_effect else ""
    eh = f" EH: {instr.exc_handler.bid}" if instr.exc_handler is not None else ""

    return f"{delta:>6} | {pc:3} {str_of_instr(instr)}{eh}"


def dump_block(
    graph: PyFlowGraph, block: Block, pc: int = 0, stack_effect: bool = False
) -> int:
    print(str_of_block_header(block))
    for instr in block.getInstructions():
        print("    ", str_of_block_instr(instr, pc, stack_effect))
        pc += graph.instrsize(instr, instr.ioparg) * opcode.CODEUNIT_SIZE
    return pc


def dump_graph(graph: PyFlowGraph, stack_effect: bool = False) -> None:
    pc = 0
    for block in graph.getBlocks():
        pc = dump_block(graph, block, pc, stack_effect)



# Portions copyright (c) Meta Platforms, Inc. and affiliates.
# pyre-strict


MANGLE_LEN = 256  # magic constant from compile.c


def mangle(name: str, klass: str | None) -> str:
    if klass is None:
        return name
    if not name.startswith("__"):
        return name
    if len(name) + 2 >= MANGLE_LEN:
        return name
    # TODO: Probably need to split and mangle recursively?
    if "." in name:
        return name
    if name.endswith("__"):
        return name
    try:
        i = 0
        while klass[i] == "_":
            i = i + 1
    except IndexError:
        return name
    klass = klass[i:]

    tlen = len(klass) + len(name)
    if tlen > MANGLE_LEN:
        klass = klass[: MANGLE_LEN - tlen]

    return "_{}{}".format(klass, name)



# Copyright (c) Meta Platforms, Inc. and affiliates.

# pyre-strict

from __future__ import annotations

import ast
import builtins
import sys
from types import CodeType
from typing import Any

from ..pyassem import (
    PyFlowGraph312,
    PyFlowGraph314,
    PyFlowGraph315,
    PyFlowGraphCinder310,
)
from ..pycodegen import (
    CinderCodeGenerator310,
    CinderCodeGenerator312,
    CinderCodeGenerator314,
    CinderCodeGenerator315,
)
from .code_gen_base import StrictCodeGenBase


# Unused but still present until we remove it from IGSRV
enable_strict_features: bool = True


class StrictCodeGenerator310(StrictCodeGenBase, CinderCodeGenerator310):
    flow_graph = PyFlowGraphCinder310


class StrictCodeGenerator312(StrictCodeGenBase, CinderCodeGenerator312):
    flow_graph = PyFlowGraph312


class StrictCodeGenerator314(StrictCodeGenerator312, CinderCodeGenerator314):
    flow_graph = PyFlowGraph314


class StrictCodeGenerator315(StrictCodeGenerator312, CinderCodeGenerator315):
    flow_graph = PyFlowGraph315


def strict_compile(
    name: str,
    filename: str,
    tree: ast.Module,
    source: str | bytes,
    optimize: int = 0,
    builtins: dict[str, Any] = builtins.__dict__,
) -> CodeType:
    code_gen = StrictCodeGenerator.make_code_gen(
        name, tree, filename, source, flags=0, optimize=optimize, builtins=builtins
    )
    return code_gen.getCode()


if sys.version_info >= (3, 15):
    StrictCodeGenerator = StrictCodeGenerator315
elif sys.version_info >= (3, 14):
    StrictCodeGenerator = StrictCodeGenerator314
elif sys.version_info >= (3, 12):
    StrictCodeGenerator = StrictCodeGenerator312
else:
    StrictCodeGenerator = StrictCodeGenerator310



# Copyright (c) Meta Platforms, Inc. and affiliates.
# pyre-strict

from .opcodebase import Opcode
from .opcodes import opcode as base_opcode


opcode: Opcode = base_opcode.copy()
opcode.def_op("LOAD_METHOD_SUPER", 198)
opcode.hasconst.add("LOAD_SUPER_METHOD")
opcode.def_op("LOAD_ATTR_SUPER", 199)
opcode.hasconst.add("LOAD_ATTR_SUPER")


opcode.stack_effects.update(
    LOAD_METHOD_SUPER=-1,
    LOAD_ATTR_SUPER=-2,
)



# Copyright (c) Meta Platforms, Inc. and affiliates.

# Generated via assign_opcode_numbers.py, do not edit.

# This is an addition to python/3.12/Lib/opcode.py, and is intended to be run
# via `exec` in generate_opcode_h.py with the globals dict obtained from
# running Lib/opcode.py.

# flake8: noqa


# Lib/opcode.py deletes these functions so we need to define them again here.
# We also need to update opname when we call def_op().
def init(
    opname,
    opmap,
    hasname,
    hasjrel,
    hasjabs,
    hasconst,
    hasarg,
    cache_format,
    specializations,
    inline_cache_entries,
    interp_only=False,
):
    def def_op(name, op):
        opmap[name] = op
        opname[op] = name

    def name_op(name, op):
        def_op(name, op)
        hasname.append(op)

    def jrel_op(name, op):
        def_op(name, op)
        hasjrel.append(op)

    def jabs_op(name, op):
        def_op(name, op)
        hasjabs.append(op)

    def_op("INVOKE_METHOD", 1)
    hasconst.append(1)
    hasarg.append(1)
    def_op("LOAD_FIELD", 2)
    hasconst.append(2)
    hasarg.append(2)
    cache_format["LOAD_FIELD"] = "{'cache': 2}"
    def_op("LOAD_OBJ_FIELD", 4)
    hasconst.append(4)
    hasarg.append(4)
    if "LOAD_FIELD" not in specializations:
        specializations["LOAD_FIELD"] = []
    specializations["LOAD_FIELD"].append("LOAD_OBJ_FIELD")
    def_op("LOAD_PRIMITIVE_FIELD", 5)
    hasconst.append(5)
    hasarg.append(5)
    if "LOAD_FIELD" not in specializations:
        specializations["LOAD_FIELD"] = []
    specializations["LOAD_FIELD"].append("LOAD_PRIMITIVE_FIELD")
    def_op("STORE_FIELD", 6)
    hasconst.append(6)
    hasarg.append(6)
    cache_format["STORE_FIELD"] = "{'cache': 2}"
    def_op("STORE_OBJ_FIELD", 7)
    hasconst.append(7)
    hasarg.append(7)
    if "STORE_FIELD" not in specializations:
        specializations["STORE_FIELD"] = []
    specializations["STORE_FIELD"].append("STORE_OBJ_FIELD")
    def_op("STORE_PRIMITIVE_FIELD", 8)
    hasconst.append(8)
    hasarg.append(8)
    if "STORE_FIELD" not in specializations:
        specializations["STORE_FIELD"] = []
    specializations["STORE_FIELD"].append("STORE_PRIMITIVE_FIELD")
    def_op("BUILD_CHECKED_LIST", 9)
    hasconst.append(9)
    hasarg.append(9)
    cache_format["BUILD_CHECKED_LIST"] = "{'cache': 2}"
    def_op("BUILD_CHECKED_LIST_CACHED", 10)
    hasconst.append(10)
    hasarg.append(10)
    if "BUILD_CHECKED_LIST" not in specializations:
        specializations["BUILD_CHECKED_LIST"] = []
    specializations["BUILD_CHECKED_LIST"].append("BUILD_CHECKED_LIST_CACHED")
    def_op("LOAD_TYPE", 11)
    hasconst.append(11)
    hasarg.append(11)
    def_op("CAST", 12)
    hasconst.append(12)
    hasarg.append(12)
    cache_format["CAST"] = "{'cache': 2}"
    def_op("CAST_CACHED", 13)
    hasconst.append(13)
    hasarg.append(13)
    if "CAST" not in specializations:
        specializations["CAST"] = []
    specializations["CAST"].append("CAST_CACHED")
    def_op("LOAD_LOCAL", 14)
    hasconst.append(14)
    hasarg.append(14)
    def_op("STORE_LOCAL", 15)
    hasconst.append(15)
    hasarg.append(15)
    cache_format["STORE_LOCAL"] = "{'cache': 1}"
    def_op("STORE_LOCAL_CACHED", 16)
    hasconst.append(16)
    hasarg.append(16)
    if "STORE_LOCAL" not in specializations:
        specializations["STORE_LOCAL"] = []
    specializations["STORE_LOCAL"].append("STORE_LOCAL_CACHED")
    def_op("PRIMITIVE_BOX", 17)
    hasarg.append(17)
    jrel_op("POP_JUMP_IF_ZERO", 100)
    hasarg.append(100)
    inline_cache_entries["POP_JUMP_IF_ZERO"] = 1
    jrel_op("POP_JUMP_IF_NONZERO", 103)
    hasarg.append(103)
    inline_cache_entries["POP_JUMP_IF_NONZERO"] = 1
    def_op("PRIMITIVE_UNBOX", 18)
    hasarg.append(18)
    def_op("PRIMITIVE_BINARY_OP", 19)
    hasarg.append(19)
    def_op("PRIMITIVE_UNARY_OP", 20)
    hasarg.append(20)
    def_op("PRIMITIVE_COMPARE_OP", 21)
    hasarg.append(21)
    def_op("LOAD_ITERABLE_ARG", 22)
    hasarg.append(22)
    def_op("LOAD_MAPPING_ARG", 23)
    hasarg.append(23)
    def_op("INVOKE_FUNCTION", 24)
    hasconst.append(24)
    hasarg.append(24)
    cache_format["INVOKE_FUNCTION"] = "{'cache': 4}"
    def_op("INVOKE_FUNCTION_CACHED", 25)
    hasconst.append(25)
    hasarg.append(25)
    if "INVOKE_FUNCTION" not in specializations:
        specializations["INVOKE_FUNCTION"] = []
    specializations["INVOKE_FUNCTION"].append("INVOKE_FUNCTION_CACHED")
    def_op("INVOKE_INDIRECT_CACHED", 26)
    hasconst.append(26)
    hasarg.append(26)
    if "INVOKE_FUNCTION" not in specializations:
        specializations["INVOKE_FUNCTION"] = []
    specializations["INVOKE_FUNCTION"].append("INVOKE_INDIRECT_CACHED")
    def_op("FAST_LEN", 27)
    hasarg.append(27)
    def_op("CONVERT_PRIMITIVE", 28)
    hasarg.append(28)
    def_op("INVOKE_NATIVE", 29)
    hasconst.append(29)
    hasarg.append(29)
    def_op("LOAD_CLASS", 30)
    hasconst.append(30)
    hasarg.append(30)
    def_op("BUILD_CHECKED_MAP", 31)
    hasconst.append(31)
    hasarg.append(31)
    cache_format["BUILD_CHECKED_MAP"] = "{'cache': 2}"
    def_op("BUILD_CHECKED_MAP_CACHED", 32)
    hasconst.append(32)
    hasarg.append(32)
    if "BUILD_CHECKED_MAP" not in specializations:
        specializations["BUILD_CHECKED_MAP"] = []
    specializations["BUILD_CHECKED_MAP"].append("BUILD_CHECKED_MAP_CACHED")
    def_op("SEQUENCE_GET", 33)
    hasarg.append(33)
    def_op("SEQUENCE_SET", 34)
    hasarg.append(34)
    def_op("LIST_DEL", 35)
    def_op("REFINE_TYPE", 36)
    hasconst.append(36)
    hasarg.append(36)
    def_op("PRIMITIVE_LOAD_CONST", 37)
    hasconst.append(37)
    hasarg.append(37)
    def_op("RETURN_PRIMITIVE", 40)
    hasarg.append(40)
    def_op("TP_ALLOC", 41)
    hasconst.append(41)
    hasarg.append(41)
    cache_format["TP_ALLOC"] = "{'cache': 2}"
    def_op("TP_ALLOC_CACHED", 42)
    hasconst.append(42)
    hasarg.append(42)
    if "TP_ALLOC" not in specializations:
        specializations["TP_ALLOC"] = []
    specializations["TP_ALLOC"].append("TP_ALLOC_CACHED")
    def_op("LOAD_METHOD_STATIC", 43)
    hasconst.append(43)
    hasarg.append(43)
    cache_format["LOAD_METHOD_STATIC"] = "{'cache': 2}"
    def_op("LOAD_METHOD_STATIC_CACHED", 45)
    hasconst.append(45)
    hasarg.append(45)
    if "LOAD_METHOD_STATIC" not in specializations:
        specializations["LOAD_METHOD_STATIC"] = []
    specializations["LOAD_METHOD_STATIC"].append("LOAD_METHOD_STATIC_CACHED")



# Copyright (c) Meta Platforms, Inc. and affiliates.
# pyre-strict

import _imp
import ast
import importlib
import os
import sys
import zipimport
from importlib.machinery import (
    BYTECODE_SUFFIXES,
    ExtensionFileLoader,
    FileFinder,
    SOURCE_SUFFIXES,
    SourceFileLoader,
    SourcelessFileLoader,
)
from types import CodeType

from .pycodegen import compile_code


# pyre-fixme[13]: path inherited but not initialized
class PySourceFileLoader(SourceFileLoader):
    # pyre-fixme[14]: source_to_code invalid override
    def source_to_code(
        self,
        data: ast.Expression | ast.Interactive | ast.Module | str,
        path: os.PathLike[str] | str,
        *,
        _optimize: int | list[int] = -1,
    ) -> CodeType:
        """Similar to SourceFileLoader.source_to_code
        but use the python based bytecode generator from
        compiler/pycodegen.py
        """
        # pyre-ignore[16]
        return importlib._bootstrap._call_with_frames_removed(
            compile_code, data, path, "exec", optimize=_optimize
        )


def _install_source_loader_helper(source_loader_type: type[PySourceFileLoader]) -> None:
    extensions = ExtensionFileLoader, _imp.extension_suffixes()
    source = source_loader_type, SOURCE_SUFFIXES
    bytecode = SourcelessFileLoader, BYTECODE_SUFFIXES
    supported_loaders = [extensions, source, bytecode]
    sys.path_hooks[:] = [
        zipimport.zipimporter,
        FileFinder.path_hook(*supported_loaders),
    ]
    sys.path_importer_cache.clear()


def _install_py_loader() -> None:
    _install_source_loader_helper(PySourceFileLoader)


def _install_strict_loader(enable_patching: bool) -> None:
    from .strict.loader import install

    install(enable_patching)



# Copyright (c) Meta Platforms, Inc. and affiliates.

# pyre-strict

from __future__ import annotations

from ast import AST, Name


def is_mutable(node: AST) -> bool:
    return isinstance(node, Name) and node.id == "mutable"



# Portions copyright (c) Meta Platforms, Inc. and affiliates.
# pyre-strict

import ast
import sys
import warnings
from collections.abc import Callable
from typing import Optional


PR_TUPLE = 0
PR_TEST = 1  # 'if'-'else', 'lambda'
PR_OR = 2  # 'or'
PR_AND = 3  # 'and'
PR_NOT = 4  # 'not'
PR_CMP = 5  # '<', '>', '==', '>=', '<=', '!=' 'in', 'not in', 'is', 'is not'
PR_EXPR = 6
PR_BOR: int = PR_EXPR  # '|'
PR_BXOR = 7  # '^'
PR_BAND = 8  # '&'
PR_SHIFT = 9  # '<<', '>>'
PR_ARITH = 10  # '+', '-'
PR_TERM = 11  # '*', '@', '/', '%', '//'
PR_FACTOR = 12  # unary '+', '-', '~'
PR_POWER = 13  # '**'
PR_AWAIT = 14  # 'await'
PR_ATOM = 15


def get_op(node: ast.cmpop) -> str:
    if isinstance(node, ast.Is):
        return " is "
    elif isinstance(node, ast.IsNot):
        return " is not "
    elif isinstance(node, ast.In):
        return " in "
    elif isinstance(node, ast.NotIn):
        return " not in "
    elif isinstance(node, ast.Lt):
        return " < "
    elif isinstance(node, ast.Gt):
        return " > "
    elif isinstance(node, ast.LtE):
        return " <= "
    elif isinstance(node, ast.GtE):
        return " >= "
    elif isinstance(node, ast.Eq):
        return " == "
    elif isinstance(node, ast.NotEq):
        return " != "
    else:
        return "unknown op: " + type(node).__name__


def get_unop(node: ast.unaryop) -> str:
    if isinstance(node, ast.UAdd):
        return "+"
    elif isinstance(node, ast.USub):
        return "-"
    elif isinstance(node, ast.Not):
        return "not "
    elif isinstance(node, ast.Invert):
        return "~"
    return "<unknown unary op>"


def _format_name(node: ast.Name, level: int) -> str:
    return node.id


def _format_compare(node: ast.Compare, level: int) -> str:
    return parens(
        level,
        PR_CMP,
        to_expr(node.left, PR_CMP + 1)
        + "".join(
            (
                get_op(op) + to_expr(comp, PR_CMP + 1)
                for comp, op in zip(node.comparators, node.ops)
            )
        ),
    )


def _format_attribute(node: ast.Attribute, level: int) -> str:
    value = to_expr(node.value, PR_ATOM)
    const = node.value
    if isinstance(const, ast.Constant) and isinstance(const.value, int):
        value += " ."
    else:
        value += "."
    return value + node.attr


def _format_tuple(node: ast.Tuple, level: int) -> str:
    if not node.elts:
        return "()"
    elif len(node.elts) == 1:
        return parens(level, PR_TUPLE, to_expr(node.elts[0]) + ",")

    return parens(level, PR_TUPLE, ", ".join(to_expr(elm) for elm in node.elts))


def _format_list(node: ast.List, level: int) -> str:
    return "[" + ", ".join(to_expr(elm) for elm in node.elts) + "]"


def _format_kw(node: ast.keyword) -> str:
    if node.arg:
        return f"{node.arg}={to_expr(node.value)}"
    return f"**{to_expr(node.value)}"


def _format_call(node: ast.Call, level: int) -> str:
    args = [to_expr(arg) for arg in node.args] + [
        _format_kw(arg) for arg in node.keywords
    ]
    return to_expr(node.func, PR_TEST) + "(" + ", ".join(args) + ")"


def _format_unaryop(node: ast.UnaryOp, level: int) -> str:
    tgt_level = PR_FACTOR
    if isinstance(node.op, ast.Not):
        tgt_level = PR_NOT
    return parens(
        level, tgt_level, get_unop(node.op) + to_expr(node.operand, tgt_level)
    )


BIN_OPS: dict[type[ast.AST], tuple[str, int]] = {
    ast.Add: (" + ", PR_ARITH),
    ast.Sub: (" - ", PR_ARITH),
    ast.Mult: (" * ", PR_TERM),
    ast.MatMult: (" @ ", PR_TERM),
    ast.Div: (" / ", PR_TERM),
    ast.Mod: (" % ", PR_TERM),
    ast.LShift: (" << ", PR_SHIFT),
    ast.RShift: (" >> ", PR_SHIFT),
    ast.BitOr: (" | ", PR_BOR),
    ast.BitXor: (" ^ ", PR_BXOR),
    ast.BitAnd: (" & ", PR_BAND),
    ast.FloorDiv: (" // ", PR_TERM),
    ast.Pow: (" ** ", PR_POWER),
}


def _format_binaryop(node: ast.BinOp, level: int) -> str:
    tgt_level = PR_FACTOR

    # pyre-fixme[6]: For 1st argument expected `Type[Union[Add, BitAnd, BitOr,
    #  BitXor, Div, FloorDiv, LShift, MatMult, Mod, Mult, Pow, RShift, Sub]]` but got
    #  `Type[operator]`.
    op, tgt_level = BIN_OPS[type(node.op)]
    rassoc = 0
    if isinstance(node.op, ast.Pow):
        rassoc = 1
    return parens(
        level,
        tgt_level,
        to_expr(node.left, tgt_level + rassoc)
        + op
        + to_expr(node.right, tgt_level + (1 - rassoc)),
    )


def _format_subscript(node: ast.Subscript, level: int) -> str:
    return f"{to_expr(node.value, PR_ATOM)}[{to_expr(node.slice, PR_TUPLE)}]"


def _format_yield(node: ast.Yield, level: int) -> str:
    raise SyntaxError("'yield expression' can not be used within an annotation")


def _format_yield_from(node: ast.YieldFrom, level: int) -> str:
    raise SyntaxError("'yield expression' can not be used within an annotation")


def _format_dict(node: ast.Dict, level: int) -> str:
    return (
        "{"
        + ", ".join(
            to_expr(k) + ": " + to_expr(v) for k, v in zip(node.keys, node.values)
        )
        + "}"
    )


def _format_comprehension(node: ast.comprehension) -> str:
    header = " for "
    if node.is_async:
        header = " async for "

    res = (
        header
        + to_expr(node.target, PR_TUPLE)
        + " in "
        + to_expr(node.iter, PR_TEST + 1)
    )
    for if_ in node.ifs:
        res += " if " + to_expr(if_, PR_TEST + 1)
    return res


def parens(level: int, target_lvl: int, value: str) -> str:
    if level > target_lvl:
        return f"({value})"
    return value


def _format_await(node: ast.Await, level: int) -> str:
    return parens(level, PR_AWAIT, "await " + to_expr(node.value, PR_ATOM))


def _format_starred(node: ast.Starred, level: int) -> str:
    return "*" + to_expr(node.value, PR_EXPR)


def _format_boolop(node: ast.BoolOp, level: int) -> str:
    if isinstance(node.op, ast.And):
        name = " and "
        tgt_level = PR_AND
    else:
        name = " or "
        tgt_level = PR_OR

    return parens(
        level, tgt_level, name.join(to_expr(n, tgt_level + 1) for n in node.values)
    )


def _format_arguments(node: ast.arguments) -> str:
    res = []
    for i, arg in enumerate(node.args):
        if i:
            res.append(", ")
        res.append(arg.arg)
        if i < len(node.defaults):
            res.append("=")
            res.append(to_expr(node.defaults[i]))

    if node.vararg or node.kwonlyargs:
        if node.args:
            res.append(", ")
        res.append("*")
        vararg = node.vararg
        if vararg:
            res.append(vararg.arg)

    for i, arg in enumerate(node.kwonlyargs):
        if res:
            res.append(", ")
        res.append(arg.arg)
        if i < len(node.kw_defaults) and node.kw_defaults[i]:
            res.append("=")
            res.append(to_expr(node.kw_defaults[i]))

    return "".join(res)


def _format_lambda(node: ast.Lambda, level: int) -> str:
    value = "lambda "
    if not node.args.args:
        value = "lambda"
    value += _format_arguments(node.args)
    value += ": " + to_expr(node.body, PR_TEST)

    return parens(level, PR_TEST, value)


def _format_if_exp(node: ast.IfExp, level: int) -> str:
    body = to_expr(node.body, PR_TEST + 1)
    orelse = to_expr(node.orelse, PR_TEST)
    test = to_expr(node.test, PR_TEST + 1)
    return parens(level, PR_TEST, f"{body} if {test} else {orelse}")


def _format_set(node: ast.Set, level: int) -> str:
    return "{" + ", ".join(to_expr(elt, PR_TEST) for elt in node.elts) + "}"


def _format_comprehensions(nodes: list[ast.comprehension]) -> str:
    return "".join(_format_comprehension(n) for n in nodes)


def _format_set_comp(node: ast.SetComp, level: int) -> str:
    return "{" + to_expr(node.elt) + _format_comprehensions(node.generators) + "}"


def _format_list_comp(node: ast.ListComp, level: int) -> str:
    return "[" + to_expr(node.elt) + _format_comprehensions(node.generators) + "]"


def _format_dict_comp(node: ast.DictComp, level: int) -> str:
    return (
        "{"
        + to_expr(node.key)
        + ": "
        + to_expr(node.value)
        + _format_comprehensions(node.generators)
        + "}"
    )


def _format_gen_exp(node: ast.GeneratorExp, level: int) -> str:
    return "(" + to_expr(node.elt) + _format_comprehensions(node.generators) + ")"


def format_fstring_elt(res: list[str], elt: ast.expr, is_format_spec: bool) -> None:
    if isinstance(elt, ast.Constant):
        res.append(elt.value)
    elif isinstance(elt, ast.JoinedStr):
        res.append(format_joinedstr(elt, PR_TEST, is_format_spec))
    elif isinstance(elt, ast.FormattedValue):
        expr = to_expr(elt.value, PR_TEST + 1)
        if expr.startswith("{"):
            # Expression starts with a brace, we need an extra space
            res.append("{ ")
        else:
            res.append("{")
        res.append(expr)
        conversion = elt.conversion
        if conversion is not None and conversion != -1:
            res.append("!")
            res.append(chr(conversion))
        format_spec = elt.format_spec
        if format_spec is not None:
            res.append(":")
            format_fstring_elt(res, format_spec, True)
        res.append("}")


def format_joinedstr(
    node: ast.JoinedStr, level: int, is_format_spec: bool = False
) -> str:
    res = []
    for elt in node.values:
        format_fstring_elt(res, elt, is_format_spec)

    joined = "".join(res)
    if is_format_spec:
        return joined
    return f"f{repr(joined)}"


def _format_slice(node: ast.Slice, level: int) -> str:
    res = ""
    if node.lower is not None:
        res += to_expr(node.lower)
    res += ":"
    if node.upper is not None:
        res += to_expr(node.upper)
    if node.step:
        res += ":"
        res += to_expr(node.step)
    return res


def _format_constant(node: ast.Constant, level: int) -> str:
    if node.value is Ellipsis:
        return "..."

    return repr(node.value)


with warnings.catch_warnings():
    warnings.simplefilter("ignore", category=DeprecationWarning)

    # pyre-ignore[9]: Pyre tries to union all the keys and values into concrete types.
    _FORMATTERS: dict[type[ast.AST], Callable[[ast.AST, int], str]] = {
        ast.Attribute: _format_attribute,
        ast.Await: _format_await,
        ast.BinOp: _format_binaryop,
        ast.BoolOp: _format_boolop,
        ast.Call: _format_call,
        ast.Compare: _format_compare,
        ast.Constant: _format_constant,
        ast.Dict: _format_dict,
        ast.DictComp: _format_dict_comp,
        ast.FormattedValue: None,
        ast.GeneratorExp: _format_gen_exp,
        ast.IfExp: _format_if_exp,
        ast.JoinedStr: format_joinedstr,
        ast.Lambda: _format_lambda,
        ast.List: _format_list,
        ast.ListComp: _format_list_comp,
        ast.Name: _format_name,
        ast.Set: _format_set,
        ast.SetComp: _format_set_comp,
        ast.Slice: _format_slice,
        ast.Starred: _format_starred,
        ast.Subscript: _format_subscript,
        ast.Tuple: _format_tuple,
        ast.UnaryOp: _format_unaryop,
        ast.Yield: _format_yield,
        ast.YieldFrom: _format_yield_from,
        **(
            {
                ast.Bytes: lambda node, level: repr(node.s),
                ast.Ellipsis: lambda node, level: "...",
            }
            if sys.version_info < (3, 14)
            else {}
        ),
    }


def to_expr(node: Optional[ast.AST], level: int = PR_TEST) -> str:
    if node is None:
        return ""
    formatter = _FORMATTERS.get(type(node))
    if formatter is not None:
        return formatter(node, level)

    return "<unsupported node: " + type(node).__name__ + ">"



# Copyright (c) Meta Platforms, Inc. and affiliates.

# pyre-strict

"""Module/script to byte-compile all .py files to .pyc files.

When called as a script with arguments, this compiles the directories
given as arguments recursively; the -l option prevents it from
recursing into directories.

Without arguments, if compiles all modules on sys.path, without
recursing into subdirectories.  (Even though it should do so for
packages -- for now, you'll have to deal with packages separately.)

See module py_compile for details of the actual byte-compilation.
"""

import filecmp
import importlib.util
import os
import py_compile
import struct
import sys
from functools import partial
from pathlib import Path
from typing import Generator, Pattern, Type

from cinderx.compiler.pysourceloader import PySourceFileLoader
from cinderx.compiler.strict.loader import strict_compile as strict_compile_fn

__all__ = ["compile_dir", "compile_file", "compile_path"]


def _walk_dir(dir: str, maxlevels: int, quiet: int = 0) -> Generator[str, None, None]:
    if quiet < 2 and isinstance(dir, os.PathLike):
        dir = os.fspath(dir)
    if not quiet:
        print("Listing {!r}...".format(dir))
    try:
        names = os.listdir(dir)
    except OSError:
        if quiet < 2:
            print("Can't list {!r}".format(dir))
        names = []
    names.sort()
    for name in names:
        if name == "__pycache__":
            continue
        fullname = os.path.join(dir, name)
        if not os.path.isdir(fullname):
            yield fullname
        elif (
            maxlevels > 0
            and name != os.curdir
            and name != os.pardir
            and os.path.isdir(fullname)
            and not os.path.islink(fullname)
        ):
            yield from _walk_dir(fullname, maxlevels=maxlevels - 1, quiet=quiet)


def compile_dir(
    dir: str,
    maxlevels: int | None = None,
    ddir: str | None = None,
    force: bool = False,
    rx: Pattern | None = None,
    quiet: int = 0,
    legacy: bool = False,
    optimize: int | list[int] = -1,
    workers: int = 1,
    invalidation_mode: py_compile.PycInvalidationMode | None = None,
    *,
    stripdir: str | None = None,
    prependdir: str | None = None,
    limit_sl_dest: str | None = None,
    hardlink_dupes: bool = False,
    loader_override: Type[PySourceFileLoader] | None = None,
    strict_compile: bool = False,
) -> bool:
    """Byte-compile all modules in the given directory tree.

    Arguments (only dir is required):

    dir:       the directory to byte-compile
    maxlevels: maximum recursion level (default `sys.getrecursionlimit()`)
    ddir:      the directory that will be prepended to the path to the
               file as it is compiled into each byte-code file.
    force:     if True, force compilation, even if timestamps are up-to-date
    quiet:     full output with False or 0, errors only with 1,
               no output with 2
    legacy:    if True, produce legacy pyc paths instead of PEP 3147 paths
    optimize:  int or list of optimization levels or -1 for level of
               the interpreter. Multiple levels leads to multiple compiled
               files each with one optimization level.
    workers:   maximum number of parallel workers
    invalidation_mode: how the up-to-dateness of the pyc will be checked
    stripdir:  part of path to left-strip from source file path
    prependdir: path to prepend to beginning of original file path, applied
               after stripdir
    limit_sl_dest: ignore symlinks if they are pointing outside of
                   the defined path
    hardlink_dupes: hardlink duplicated pyc files
    loader_override: loader type to use instead of default SourceFileLoader
    strict_compile: Whether to use the strict compiler instead of the default.
    """
    poolexecutor = None
    if ddir is not None and (stripdir is not None or prependdir is not None):
        raise ValueError(
            (
                "Destination dir (ddir) cannot be used "
                "in combination with stripdir or prependdir"
            )
        )
    if ddir is not None:
        stripdir = dir
        prependdir = ddir
        ddir = None
    if workers < 0:
        raise ValueError("workers must be greater or equal to 0")
    if workers != 1:
        # Check if this is a system where ProcessPoolExecutor can function.
        from concurrent.futures.process import _check_system_limits

        try:
            _check_system_limits()
        except NotImplementedError:
            workers = 1
        else:
            from concurrent.futures import ProcessPoolExecutor

            poolexecutor = ProcessPoolExecutor
    if maxlevels is None:
        maxlevels = sys.getrecursionlimit()
    files = _walk_dir(dir, quiet=quiet, maxlevels=maxlevels)
    success = True
    if workers != 1 and poolexecutor is not None:
        # If workers == 0, let ProcessPoolExecutor choose
        workers = workers or None
        with poolexecutor(max_workers=workers) as executor:
            results = executor.map(
                partial(
                    compile_file,
                    ddir=ddir,
                    force=force,
                    rx=rx,
                    quiet=quiet,
                    legacy=legacy,
                    optimize=optimize,
                    invalidation_mode=invalidation_mode,
                    stripdir=stripdir,
                    prependdir=prependdir,
                    limit_sl_dest=limit_sl_dest,
                    hardlink_dupes=hardlink_dupes,
                    loader_override=loader_override,
                    strict_compile=strict_compile,
                ),
                files,
            )
            success = min(results, default=True)
    else:
        for file in files:
            if not compile_file(
                file,
                ddir,
                force,
                rx,
                quiet,
                legacy,
                optimize,
                invalidation_mode,
                stripdir=stripdir,
                prependdir=prependdir,
                limit_sl_dest=limit_sl_dest,
                hardlink_dupes=hardlink_dupes,
                loader_override=loader_override,
                strict_compile=strict_compile,
            ):
                success = False
    return success


def compile_file(
    fullname: str,
    ddir: str | None = None,
    force: bool = False,
    rx: Pattern | None = None,
    quiet: int = 0,
    legacy: bool = False,
    optimize: int | list[int] = -1,
    invalidation_mode: py_compile.PycInvalidationMode | None = None,
    *,
    stripdir: str | None = None,
    prependdir: str | None = None,
    limit_sl_dest: str | None = None,
    hardlink_dupes: bool = False,
    loader_override: Type[PySourceFileLoader] | None = None,
    strict_compile: bool = False,
) -> bool:
    """Byte-compile one file.

    Arguments (only fullname is required):

    fullname:  the file to byte-compile
    ddir:      if given, the directory name compiled in to the
               byte-code file.
    force:     if True, force compilation, even if timestamps are up-to-date
    quiet:     full output with False or 0, errors only with 1,
               no output with 2
    legacy:    if True, produce legacy pyc paths instead of PEP 3147 paths
    optimize:  int or list of optimization levels or -1 for level of
               the interpreter. Multiple levels leads to multiple compiled
               files each with one optimization level.
    invalidation_mode: how the up-to-dateness of the pyc will be checked
    stripdir:  part of path to left-strip from source file path
    prependdir: path to prepend to beginning of original file path, applied
               after stripdir
    limit_sl_dest: ignore symlinks if they are pointing outside of
                   the defined path.
    hardlink_dupes: hardlink duplicated pyc files
    loader_override: loader type to use instead of default SourceFileLoader
    """

    if ddir is not None and (stripdir is not None or prependdir is not None):
        raise ValueError(
            (
                "Destination dir (ddir) cannot be used "
                "in combination with stripdir or prependdir"
            )
        )

    success = True
    if quiet < 2 and isinstance(fullname, os.PathLike):
        fullname = os.fspath(fullname)
    name = os.path.basename(fullname)

    dfile = None

    if ddir is not None:
        dfile = os.path.join(ddir, name)

    if stripdir is not None:
        fullname_parts = fullname.split(os.path.sep)
        stripdir_parts = stripdir.split(os.path.sep)
        ddir_parts = list(fullname_parts)

        for spart, opart in zip(stripdir_parts, fullname_parts):
            if spart == opart:
                ddir_parts.remove(spart)

        dfile = os.path.join(*ddir_parts)

    if prependdir is not None:
        if dfile is None:
            dfile = os.path.join(prependdir, fullname)
        else:
            dfile = os.path.join(prependdir, dfile)

    if isinstance(optimize, int):
        optimize = [optimize]

    # Use set() to remove duplicates.
    # Use sorted() to create pyc files in a deterministic order.
    optimize = sorted(set(optimize))

    if hardlink_dupes and len(optimize) < 2:
        raise ValueError(
            "Hardlinking of duplicated bytecode makes sense "
            "only for more than one optimization level"
        )

    if rx is not None:
        mo = rx.search(fullname)
        if mo:
            return success

    if limit_sl_dest is not None and os.path.islink(fullname):
        if Path(limit_sl_dest).resolve() not in Path(fullname).resolve().parents:
            return success

    opt_cfiles = {}

    if os.path.isfile(fullname):
        for opt_level in optimize:
            if legacy:
                opt_cfiles[opt_level] = fullname + "c"
            else:
                if opt_level >= 0:
                    opt = opt_level if opt_level >= 1 else ""
                    cfile = importlib.util.cache_from_source(fullname, optimization=opt)
                    opt_cfiles[opt_level] = cfile
                else:
                    cfile = importlib.util.cache_from_source(fullname)
                    opt_cfiles[opt_level] = cfile

        tail = name[-3:]
        if tail == ".py":
            if not force:
                try:
                    mtime = int(os.stat(fullname).st_mtime)
                    expect = struct.pack(
                        "<4sLL", importlib.util.MAGIC_NUMBER, 0, mtime & 0xFFFF_FFFF
                    )
                    for cfile in opt_cfiles.values():
                        with open(cfile, "rb") as chandle:
                            actual = chandle.read(12)
                        if expect != actual:
                            break
                    else:
                        return success
                except OSError:
                    pass
            if not quiet:
                print("Compiling {!r}...".format(fullname))
            ok = 0
            try:
                for index, opt_level in enumerate(optimize):
                    cfile = opt_cfiles[opt_level]
                    if strict_compile:
                        ok = strict_compile_fn(
                            fullname,
                            cfile,
                            dfile,
                            True,
                            optimize=opt_level,
                            invalidation_mode=invalidation_mode,
                            loader_override=loader_override,
                        )
                    else:
                        ok = py_compile.compile(
                            fullname,
                            cfile,
                            dfile,
                            True,
                            optimize=opt_level,
                            invalidation_mode=invalidation_mode,
                        )
                    if index > 0 and hardlink_dupes:
                        previous_cfile = opt_cfiles[optimize[index - 1]]
                        if filecmp.cmp(cfile, previous_cfile, shallow=False):
                            os.unlink(cfile)
                            os.link(previous_cfile, cfile)
            except py_compile.PyCompileError as err:
                success = False
                if quiet >= 2:
                    return success
                elif quiet:
                    print("*** Error compiling {!r}...".format(fullname))
                else:
                    print("*** ", end="")
                # escape non-printable characters in msg
                encoding = sys.stdout.encoding or sys.getdefaultencoding()
                msg = err.msg.encode(encoding, errors="backslashreplace").decode(
                    encoding
                )
                print(msg)
            except (SyntaxError, UnicodeError, OSError) as e:
                success = False
                if quiet >= 2:
                    return success
                elif quiet:
                    print("*** Error compiling {!r}...".format(fullname))
                else:
                    print("*** ", end="")
                print(e.__class__.__name__ + ":", e)
            else:
                if ok == 0:
                    success = False
    return success


def compile_path(
    skip_curdir: int = 1,
    maxlevels: int = 0,
    force: bool = False,
    quiet: int = 0,
    legacy: bool = False,
    optimize: int | list[int] = -1,
    invalidation_mode: py_compile.PycInvalidationMode | None = None,
    loader_override: Type[PySourceFileLoader] | None = None,
    strict_compile: bool = False,
) -> bool:
    """Byte-compile all module on sys.path.

    Arguments (all optional):

    skip_curdir: if true, skip current directory (default True)
    maxlevels:   max recursion level (default 0)
    force: as for compile_dir() (default False)
    quiet: as for compile_dir() (default 0)
    legacy: as for compile_dir() (default False)
    optimize: as for compile_dir() (default -1)
    invalidation_mode: as for compiler_dir()
    loader_override: as for compiler_dir()
    """
    success = True
    for dir in sys.path:
        if (not dir or dir == os.curdir) and skip_curdir:
            if quiet < 2:
                print("Skipping current directory")
        else:
            success = success and compile_dir(
                dir,
                maxlevels,
                None,
                force,
                quiet=quiet,
                legacy=legacy,
                optimize=optimize,
                invalidation_mode=invalidation_mode,
                loader_override=loader_override,
                strict_compile=strict_compile,
            )
    return success


def main() -> bool:
    """Script main program."""
    import argparse

    parser = argparse.ArgumentParser(
        description="Utilities to support installing Python libraries."
    )
    parser.add_argument(
        "-l",
        action="store_const",
        const=0,
        default=None,
        dest="maxlevels",
        help="don't recurse into subdirectories",
    )
    parser.add_argument(
        "-r",
        type=int,
        dest="recursion",
        help=(
            "control the maximum recursion level. "
            "if `-l` and `-r` options are specified, "
            "then `-r` takes precedence."
        ),
    )
    parser.add_argument(
        "-f",
        action="store_true",
        dest="force",
        help="force rebuild even if timestamps are up to date",
    )
    parser.add_argument(
        "-q",
        action="count",
        dest="quiet",
        default=0,
        help="output only error messages; -qq will suppress "
        "the error messages as well.",
    )
    parser.add_argument(
        "-b",
        action="store_true",
        dest="legacy",
        help="use legacy (pre-PEP3147) compiled file locations",
    )
    parser.add_argument(
        "-d",
        metavar="DESTDIR",
        dest="ddir",
        default=None,
        help=(
            "directory to prepend to file paths for use in "
            "compile-time tracebacks and in runtime "
            "tracebacks in cases where the source file is "
            "unavailable"
        ),
    )
    parser.add_argument(
        "-s",
        metavar="STRIPDIR",
        dest="stripdir",
        default=None,
        help=(
            "part of path to left-strip from path "
            "to source file - for example buildroot. "
            "`-d` and `-s` options cannot be "
            "specified together."
        ),
    )
    parser.add_argument(
        "-p",
        metavar="PREPENDDIR",
        dest="prependdir",
        default=None,
        help=(
            "path to add as prefix to path "
            "to source file - for example / to make "
            "it absolute when some part is removed "
            "by `-s` option. "
            "`-d` and `-p` options cannot be "
            "specified together."
        ),
    )
    parser.add_argument(
        "-x",
        metavar="REGEXP",
        dest="rx",
        default=None,
        help=(
            "skip files matching the regular expression; "
            "the regexp is searched for in the full path "
            "of each file considered for compilation"
        ),
    )
    parser.add_argument(
        "-i",
        metavar="FILE",
        dest="flist",
        help=(
            "add all the files and directories listed in "
            "FILE to the list considered for compilation; "
            'if "-", names are read from stdin'
        ),
    )
    parser.add_argument(
        "compile_dest",
        metavar="FILE|DIR",
        nargs="*",
        help=(
            "zero or more file and directory names "
            "to compile; if no arguments given, defaults "
            "to the equivalent of -l sys.path"
        ),
    )
    parser.add_argument(
        "-j", "--workers", default=1, type=int, help="Run compileall concurrently"
    )
    invalidation_modes = [
        mode.name.lower().replace("_", "-") for mode in py_compile.PycInvalidationMode
    ]
    parser.add_argument(
        "--invalidation-mode",
        choices=sorted(invalidation_modes),
        help=(
            "set .pyc invalidation mode; defaults to "
            '"checked-hash" if the SOURCE_DATE_EPOCH '
            "environment variable is set, and "
            '"timestamp" otherwise.'
        ),
    )
    parser.add_argument(
        "-o",
        action="append",
        type=int,
        dest="opt_levels",
        help=(
            "Optimization levels to run compilation with. "
            "Default is -1 which uses the optimization level "
            "of the Python interpreter itself (see -O)."
        ),
    )
    parser.add_argument(
        "-e",
        metavar="DIR",
        dest="limit_sl_dest",
        help="Ignore symlinks pointing outsite of the DIR",
    )
    parser.add_argument(
        "--hardlink-dupes",
        action="store_true",
        dest="hardlink_dupes",
        help="Hardlink duplicated pyc files",
    )
    parser.add_argument(
        "--python-loader",
        action="store_true",
        dest="use_py_loader",
        help=("use loader that uses CinderX compiler"),
    )
    parser.add_argument(
        "--strict-compile",
        action="store_true",
        dest="strict_compile",
        help=("use the bytecode compiler bundled with the strict-module loader"),
    )

    args = parser.parse_args()
    compile_dests = args.compile_dest

    if args.rx:
        import re

        args.rx = re.compile(args.rx)

    if args.limit_sl_dest == "":
        args.limit_sl_dest = None

    if args.recursion is not None:
        maxlevels = args.recursion
    else:
        maxlevels = args.maxlevels

    if args.opt_levels is None:
        args.opt_levels = [-1]

    if len(args.opt_levels) == 1 and args.hardlink_dupes:
        parser.error(
            (
                "Hardlinking of duplicated bytecode makes sense "
                "only for more than one optimization level."
            )
        )

    if args.ddir is not None and (
        args.stripdir is not None or args.prependdir is not None
    ):
        parser.error("-d cannot be used in combination with -s or -p")

    # if flist is provided then load it
    if args.flist:
        try:
            with (
                sys.stdin if args.flist == "-" else open(args.flist, encoding="utf-8")
            ) as f:
                for line in f:
                    compile_dests.append(line.strip())
        except OSError:
            if args.quiet < 2:
                print("Error reading file list {}".format(args.flist))
            return False

    if args.invalidation_mode:
        ivl_mode = args.invalidation_mode.replace("-", "_").upper()
        invalidation_mode = py_compile.PycInvalidationMode[ivl_mode]
    else:
        invalidation_mode = None

    success = True
    loader_override = None
    if args.use_py_loader:
        loader_override = PySourceFileLoader
        sys.setrecursionlimit(sys.getrecursionlimit() * 100)
    strict_compile = args.strict_compile
    try:
        if compile_dests:
            for dest in compile_dests:
                if os.path.isfile(dest):
                    if not compile_file(
                        dest,
                        args.ddir,
                        args.force,
                        args.rx,
                        args.quiet,
                        args.legacy,
                        invalidation_mode=invalidation_mode,
                        stripdir=args.stripdir,
                        prependdir=args.prependdir,
                        optimize=args.opt_levels,
                        limit_sl_dest=args.limit_sl_dest,
                        hardlink_dupes=args.hardlink_dupes,
                        loader_override=loader_override,
                        strict_compile=strict_compile,
                    ):
                        success = False
                else:
                    appended = False
                    try:
                        if strict_compile and dest not in sys.path:
                            sys.path.append(dest)
                            appended = True

                        if not compile_dir(
                            dest,
                            maxlevels,
                            args.ddir,
                            args.force,
                            args.rx,
                            args.quiet,
                            args.legacy,
                            workers=args.workers,
                            invalidation_mode=invalidation_mode,
                            stripdir=args.stripdir,
                            prependdir=args.prependdir,
                            optimize=args.opt_levels,
                            limit_sl_dest=args.limit_sl_dest,
                            hardlink_dupes=args.hardlink_dupes,
                            loader_override=loader_override,
                            strict_compile=strict_compile,
                        ):
                            success = False
                    finally:
                        if appended:
                            sys.path.remove(dest)
            return success
        else:
            return compile_path(
                legacy=args.legacy,
                force=args.force,
                quiet=args.quiet,
                invalidation_mode=invalidation_mode,
                loader_override=loader_override,
                strict_compile=strict_compile,
            )
    except KeyboardInterrupt:
        if args.quiet < 2:
            print("\n[interrupted]")
        return False
    return True


if __name__ == "__main__":
    exit_status = int(not main())
    sys.exit(exit_status)



# Copyright (c) Meta Platforms, Inc. and affiliates.

# pyre-strict

from __future__ import annotations

import ast
from ast import AST, Attribute, BinOp, Call, ClassDef, Constant, Name, Subscript
from contextlib import nullcontext
from enum import Enum
from typing import cast, ContextManager, Optional, TYPE_CHECKING

from ..errors import TypedSyntaxError
from ..symbols import ModuleScope, Scope
from .types import (
    Callable,
    Class,
    ClassVar,
    CType,
    DataclassDecorator,
    DynamicClass,
    ExactClass,
    FinalClass,
    Function,
    FunctionGroup,
    InitVar,
    KnownBoolean,
    MethodType,
    ModuleInstance,
    NativeDecorator,
    TType,
    TypeDescr,
    UnionType,
    UnknownDecoratedMethod,
    Value,
)
from .visitor import GenericVisitor

if TYPE_CHECKING:
    from .compiler import Compiler


# When set to True, allow type alises of the form
#   A = Foo
# as well as the explicit 3.12+
#   type A = Foo
# When set to False, only the latter form will be supported.
#
# TASK(T214139735): This is a placeholder for a feature flag that will let
# users opt in to implicit type aliases on a per-module basis.
ENABLE_IMPLICIT_TYPE_ALIASES = False


class ModuleFlag(Enum):
    CHECKED_DICTS = 1
    CHECKED_LISTS = 3


class ModuleTableException(Exception):
    pass


class ReferenceVisitor(GenericVisitor[Optional[Value]]):
    def __init__(self, module: ModuleTable) -> None:
        super().__init__(module)
        self.subscr_nesting = 0
        self.local_names: dict[str, Value] = {}

    def visitName(self, node: Name) -> Value | None:
        if node.id in self.local_names:
            return self.local_names[node.id]
        return self.module.get_child(
            node.id, self.context_qualname, force_decl=self.force_decl_deps
        ) or self.module.compiler.builtins.get_child_intrinsic(node.id)

    def visitAttribute(self, node: Attribute) -> Value | None:
        val = self.visit(node.value)
        if val is not None:
            return val.resolve_attr(node, self)

    def add_local_name(self, name: str, value: Value) -> None:
        self.local_names[name] = value

    def clear_local_names(self) -> None:
        self.local_names = {}


class AnnotationVisitor(ReferenceVisitor):
    def resolve_annotation(
        self,
        node: ast.AST,
        *,
        is_declaration: bool = False,
    ) -> Class | None:
        with self.error_context(node):
            klass = self.visit(node)
            if not isinstance(klass, Class):
                return None

            if self.subscr_nesting or not is_declaration:
                if isinstance(klass, FinalClass):
                    raise TypedSyntaxError(
                        "Final annotation is only valid in initial declaration "
                        "of attribute or module-level constant",
                    )
                if isinstance(klass, ClassVar):
                    raise TypedSyntaxError(
                        "ClassVar is allowed only in class attribute annotations. "
                        "Class Finals are inferred ClassVar; do not nest with Final."
                    )
                if isinstance(klass, InitVar):
                    raise TypedSyntaxError(
                        "InitVar is allowed only in class attribute annotations."
                    )

            if isinstance(klass, ExactClass):
                klass = klass.unwrap().exact_type()
            elif isinstance(klass, FinalClass):
                pass
            else:
                klass = klass.inexact_type()
            # PEP-484 specifies that ints should be treated as a subclass of floats,
            # even though they differ in the runtime. We need to maintain the distinction
            # between the two internally, so we should view user-specified `float` annotations
            # as `float | int`. This widening of the type prevents us from applying
            # optimizations to user-specified floats, but does not affect ints. Since we
            # don't optimize Python floats anyway, we accept this to maintain PEP-484 compatibility.

            if klass.unwrap() is self.type_env.float:
                klass = self.compiler.type_env.get_union(
                    (self.type_env.float, self.type_env.int)
                )

            # TODO until we support runtime checking of unions, we must for
            # safety resolve union annotations to dynamic (except for
            # optionals, which we can check at runtime)
            if self.is_unsupported_union_type(klass):
                return None

            # Types like typing.NamedTuple or typing.Protocol should not be
            # used as annotations; if they are, fall back to dynamic rather
            # than raise an error (see T186572841 for context).

            # TASK(T186572841): Extend to Protocol as well
            special_types = ("NamedTuple", "TypedDict")
            if klass.type_name.module == "typing":
                if klass.type_name.qualname in special_types:
                    return self.type_env.dynamic

            return klass

    def is_unsupported_union_type(self, klass: Value) -> bool:
        """Returns True if klass is an unsupported union type."""
        return (
            isinstance(klass, UnionType)
            and klass is not self.type_env.union
            and klass is not self.type_env.optional
            and klass.opt_type is None
        )

    def visitSubscript(self, node: Subscript) -> Value | None:
        target = self.resolve_annotation(node.value, is_declaration=True)
        if target is None:
            return None

        self.subscr_nesting += 1
        slice = self.visit(node.slice) or self.type_env.DYNAMIC
        self.subscr_nesting -= 1
        return target.resolve_subscr(node, slice, self) or target

    def visitBinOp(self, node: BinOp) -> Value | None:
        if isinstance(node.op, ast.BitOr):
            ltype = self.resolve_annotation(node.left)
            rtype = self.resolve_annotation(node.right)
            if ltype is None or rtype is None:
                return None
            return self.module.compiler.type_env.get_union((ltype, rtype))

    def visitConstant(self, node: Constant) -> Value | None:
        sval = node.value
        if sval is None:
            return self.type_env.none
        elif isinstance(sval, str):
            n = ast.parse(node.value, "", "eval").body
            return self.visit(n)


class DepTrackingOptOut:
    """A reason for opting out of module dependency tracking.

    This helps ensure our dependency tracking is complete; we can't request type
    information from a module via ModuleTable.get_child(...) without either
    providing a `requester` (to track the dependency) or providing an opt-out
    reason.
    """

    def __init__(self, reason: str) -> None:
        self.reason = reason


INTRINSIC_OPT_OUT = DepTrackingOptOut(
    "We don't need to track dependencies to intrinsic modules; "
    "they won't change during development and require pyc updates. "
    "If they change, we should bump the static pyc magic number."
)
DEFERRED_IMPORT_OPT_OUT = DepTrackingOptOut("Tracked in ModuleTable.declare_import()")


class DeferredImport:
    def __init__(
        self,
        in_module: ModuleTable,
        mod_to_import: str,
        asname: str,
        optimize: int,
        compiler: Compiler,
        name: str | None = None,
        mod_to_return: str | None = None,
    ) -> None:
        self.in_module = in_module
        self.mod_to_import = mod_to_import
        self.asname = asname
        self.name = name
        self.mod_to_return = mod_to_return
        self.compiler = compiler
        self.optimize = optimize

    def import_mod(self, mod_name: str) -> ModuleTable | None:
        return self.compiler.import_module(mod_name, optimize=self.optimize)

    def resolve(self) -> Value:
        if self.mod_to_import not in self.compiler.modules:
            self.import_mod(self.mod_to_import)
        mod = self.compiler.modules.get(self.mod_to_import)
        if mod is not None:
            if self.name is None:
                return ModuleInstance(
                    self.mod_to_return or self.mod_to_import, self.compiler
                )
            val = mod.get_child(self.name, DEFERRED_IMPORT_OPT_OUT)
            if val is not None:
                return val
            try_mod = f"{self.mod_to_import}.{self.name}"
            self.import_mod(try_mod)
            if try_mod in self.compiler.modules:
                return ModuleInstance(try_mod, self.compiler)
            if not mod.first_pass_done:
                raise ModuleTableException(
                    f"Cannot find {self.mod_to_import}.{self.name} due to cyclic reference"
                )
        # For unknown modules, we don't need to record each individual object
        # used from them; it doesn't matter, as we won't track transitive deps
        # across them. We just need to record a dependency to the module in
        # general.
        self.in_module.record_dependency(self.asname, (self.mod_to_import, "<any>"))
        return self.compiler.type_env.DYNAMIC


class UnknownModule:
    def __init__(self, name: str) -> None:
        self.name = name


class ModuleTable:
    def __init__(
        self,
        name: str,
        filename: str,
        compiler: Compiler,
        members: dict[str, Value] | None = None,
        first_pass_done: bool = True,
    ) -> None:
        self.name = name
        self.filename = filename
        self._children: dict[str, Value | DeferredImport] = {}
        if members is not None:
            self._children.update(members)
        self.compiler = compiler
        self.types: dict[AST, Value] = {}
        self.node_data: dict[tuple[AST, object], object] = {}
        self.flags: set[ModuleFlag] = set()
        self.decls: list[tuple[AST, str | None, Value | None]] = []
        self.implicit_decl_names: set[str] = set()
        self.type_alias_names: set[str] = set()
        self.compile_non_static: set[AST] = set()
        # {local-name: {(mod, qualname)}} for decl-time deps
        self.decl_deps: dict[str, set[tuple[str, str]]] = {}
        # {local-name: {(mod, qualname)}} for bind-time deps
        self.bind_deps: dict[str, set[tuple[str, str]]] = {}
        # (source module, source name) for every name this module imports-from
        # another static module at top level
        self.imported_from: dict[str, tuple[str, str]] = {}
        # TODO: final constants should be typed to literals, and
        # this should be removed in the future
        self.named_finals: dict[str, ast.Constant] = {}
        # Have we completed our first pass through the module, populating
        # imports and types defined in the module? Until we have, resolving
        # type annotations is not safe.
        self.first_pass_done = first_pass_done
        self.finish_bind_done = False
        self.ann_visitor = AnnotationVisitor(self)
        self.ref_visitor = ReferenceVisitor(self)
        # the prefix children will get on their qualname; always None for
        # modules (the qualname of a class or function is within-module, it
        # doesn't include the module name)
        self.qualname: None = None

    def __repr__(self) -> str:
        return f"<ModuleTable {self.name}>"

    def get_dependencies(self) -> set[ModuleTable | UnknownModule]:
        """Return all modules this module depends on."""
        # We only include bind deps for the current module; for transitive deps
        # we follow only decl_deps
        immediate_deps = {**self.bind_deps, **self.decl_deps}
        all_modules_deps = {
            name: mod.decl_deps for name, mod in self.compiler.modules.items()
        }
        all_modules_deps[self.name] = immediate_deps
        dep_names = find_transitive_deps(self.name, all_modules_deps)
        return {
            self.compiler.modules.get(name) or UnknownModule(name) for name in dep_names
        }

    def record_dependency(
        self, name: str, target: tuple[str, str], force_decl: bool = False
    ) -> None:
        """Record a dependency from a local name to a name in another module.

        `name` is the name in this module's namespace. `target` is a (str, str)
        tuple of (source_module, source_name).

        """
        deps = (
            self.bind_deps
            if (self.finish_bind_done and not force_decl)
            else self.decl_deps
        )
        deps.setdefault(name, set()).add(target)

    def get_child(
        self, name: str, requester: str | DepTrackingOptOut, force_decl: bool = False
    ) -> Value | None:
        if not isinstance(requester, DepTrackingOptOut):
            # Using imported_from here is just an optimization that lets us skip
            # one level of transitive-dependency-following later. If modA has
            # "from modB import B", we already record `modA.B -> modB.B`, so
            # technically if `modA.f -> modA.B`, we could just record that
            # dependency, and following the transitive closure would later give
            # us `modA.f -> modB.B`. But since we have the `imported_from` data
            # available, we use it to just record `modA.f -> modB.B` initially.
            # We don't need `modA.f -> modA.B` recorded as a dependency in that
            # case, since intra-module dependencies have no direct impact on
            # recompilation, they are only needed for following the transitive
            # chain across modules.
            source = self.imported_from.get(name, (self.name, name))
            self.record_dependency(requester, source, force_decl)
        res = self._children.get(name)
        if isinstance(res, DeferredImport):
            self._children[name] = res = res.resolve()
        return res

    def syntax_error(self, msg: str, node: AST) -> None:
        return self.compiler.error_sink.syntax_error(msg, self.filename, node)

    def error_context(self, node: AST | None) -> ContextManager[None]:
        if node is None:
            return nullcontext()
        return self.compiler.error_sink.error_context(self.filename, node)

    def maybe_set_type_alias(
        self,
        # pyre-ignore[11]: Annotation `ast.TypeAlias` is not defined as a type
        node: ast.Assign | ast.TypeAlias,
        name: str,
        *,
        require_type: bool = False,
    ) -> None:
        """
        Check if we are assigning a Class or Union value to a variable at
        module scope, and if so, store it as a type alias.
        """
        value = self.resolve_type(node.value, name)
        if isinstance(value, Class):
            # This is a type, treat it similarly to a class declaration
            self.decls.append((node, name, value))
            self._children[name] = value
            self.type_alias_names.add(name)
        else:
            # Treat the type as dynamic if it is an assignment,
            # raise an error if it is an explicit type alias.
            if require_type:
                rhs = ast.unparse(node.value)
                raise TypedSyntaxError(f"RHS of type alias {name} is not a type: {rhs}")
            self.implicit_decl_names.add(name)

    # pyre-ignore[11]: Annotation `ast.TypeAlias` is not defined as a type
    def declare_type_alias(self, node: ast.TypeAlias) -> None:
        self.maybe_set_type_alias(node, node.name.id, require_type=True)

    def declare_class(self, node: ClassDef, klass: Class) -> None:
        if self.first_pass_done:
            raise ModuleTableException(
                "Attempted to declare a class after the declaration visit"
            )
        self.decls.append((node, node.name, klass))
        self._children[node.name] = klass

    def declare_function(self, func: Function) -> None:
        if self.first_pass_done:
            raise ModuleTableException(
                "Attempted to declare a function after the declaration visit"
            )
        existing = self._children.get(func.func_name)
        new_member = func
        if existing is not None:
            if isinstance(existing, Function):
                new_member = FunctionGroup([existing, new_member], func.klass.type_env)
            elif isinstance(existing, FunctionGroup):
                existing.functions.append(new_member)
                new_member = existing
            else:
                raise TypedSyntaxError(
                    f"function conflicts with other member {func.func_name} in {self.name}"
                )

        self.decls.append((func.node, func.func_name, new_member))
        self._children[func.func_name] = new_member

    def _get_inferred_type(self, value: ast.expr, requester: str) -> Value | None:
        if not isinstance(value, ast.Name):
            return None
        return self.get_child(value.id, requester)

    def finish_bind(self) -> None:
        self.first_pass_done = True
        for node, name, value in self.decls:
            with self.error_context(node):
                if value is not None:
                    assert name is not None
                    new_value = value.finish_bind(self, None)
                    if new_value is None:
                        # We still need to maintain a name for unknown decorated methods since they
                        # will bind to some name, and we want to avoid throwing unknown name errors.
                        if isinstance(self.types[node], UnknownDecoratedMethod):
                            self._children[name] = self.compiler.type_env.DYNAMIC
                        else:
                            del self._children[name]
                    elif new_value is not value:
                        self._children[name] = new_value

                if isinstance(node, ast.AnnAssign) and isinstance(
                    node.target, ast.Name
                ):
                    # isinstance(target := node.target, ast.Name) above would be
                    # nicer, but pyre doesn't narrow the type of target :/
                    target = node.target
                    assert isinstance(target, ast.Name)
                    typ = self.resolve_annotation(
                        node.annotation, target.id, is_declaration=True
                    )
                    if typ is not None:
                        # Special case Final[dynamic] to use inferred type.
                        instance = typ.instance
                        value = node.value
                        if isinstance(typ, FinalClass):
                            # We keep track of annotated finals in the
                            # named_finals field - we can safely remove that
                            # information here to ensure that the rest of the
                            # type system can safely ignore it.
                            unwrapped = typ.unwrap()
                            if value is not None and isinstance(
                                unwrapped, DynamicClass
                            ):
                                instance = (
                                    self._get_inferred_type(value, target.id)
                                    or unwrapped.instance
                                )
                            else:
                                instance = unwrapped.instance

                        self._children[target.id] = instance

                    if isinstance(typ, FinalClass):
                        value = node.value
                        if not value:
                            raise TypedSyntaxError(
                                "Must assign a value when declaring a Final"
                            )
                        elif not isinstance(typ, CType) and isinstance(
                            value, ast.Constant
                        ):
                            self.named_finals[target.id] = value

        for name in self.implicit_decl_names:
            if name not in self._children:
                self._children[name] = self.compiler.type_env.DYNAMIC
        # We don't need these anymore...
        self.implicit_decl_names.clear()
        self.finish_bind_done = True

    def validate_overrides(self) -> None:
        for _node, name, _value in self.decls:
            if name is None or name in self.type_alias_names:
                continue

            child = self._children.get(name, None)
            if isinstance(child, Value):
                child.validate_overrides(self, None)

        self.decls.clear()

    def resolve_type(self, node: ast.AST, requester: str) -> Class | None:
        with self.ann_visitor.temporary_context_qualname(requester):
            try:
                typ = self.ann_visitor.visit(node)
            except SyntaxError:
                # We resolve types via the AnnotationVisitor, which expects
                # string literals to parse as well-formed python type
                # expressions. In the more general case here (where we are
                # resolving a node that is not necessarily an annotation in the
                # source code) the visitor will try to parse arbitrary string
                # literals as python code, and will raise a SyntaxError.
                return None
        if isinstance(typ, Class):
            return typ

    def resolve_decorator(self, node: ast.AST) -> Value | None:
        if isinstance(node, Call):
            func = self.ref_visitor.visit(node.func)
            if isinstance(func, Class):
                return func.instance
            elif isinstance(func, DataclassDecorator):
                return func
            elif isinstance(func, NativeDecorator):
                return func
            elif isinstance(func, Callable):
                return func.return_type.resolved().instance
            elif isinstance(func, MethodType):
                return func.function.return_type.resolved().instance

        return self.ref_visitor.visit(node)

    def resolve_annotation(
        self,
        node: ast.AST,
        requester: str,
        *,
        # annotation on a variable or attribute declaration (could be inside a
        # function, thus not public)
        is_declaration: bool = False,
        # annotation on public API (also includes e.g. function arg/return
        # annotations, does not include function-internal declarations)
        is_decl_dep: bool = False,
    ) -> Class | None:
        with self.ann_visitor.temporary_context_qualname(requester, is_decl_dep):
            return self.ann_visitor.resolve_annotation(
                node, is_declaration=is_declaration
            )

    def resolve_name_with_descr(
        self, name: str, requester: str
    ) -> tuple[Value | None, TypeDescr | None]:
        if val := self.get_child(name, requester):
            return val, ((self.name,), name)
        elif val := self.compiler.builtins.get_child_intrinsic(name):
            return val, None
        return None, None

    def get_final_literal(self, node: AST, scope: Scope) -> ast.Constant | None:
        if not isinstance(node, Name):
            return None

        final_val = self.named_finals.get(node.id, None)
        if (
            final_val is not None
            and isinstance(node.ctx, ast.Load)
            and (
                # Ensure the name is not shadowed in the local scope
                isinstance(scope, ModuleScope) or node.id not in scope.defs
            )
        ):
            return final_val

    def declare_import(
        self, name: str, target: tuple[str, str] | None, val: Value | DeferredImport
    ) -> None:
        """Declare a name imported into this module.

        `name` is the name in this module's namespace. `target` is a (str, str)
        tuple of (source_module, source_name) for an `import from`. For a
        top-level module import, `target` should be `None`.

        """
        if self.first_pass_done:
            raise ModuleTableException(
                "Attempted to declare an import after the declaration visit"
            )
        self._children[name] = val
        if target is not None:
            self.imported_from[name] = target
            self.record_dependency(name, target)

    def declare_variable(self, node: ast.AnnAssign, module: ModuleTable) -> None:
        self.decls.append((node, None, None))

    def declare_variables(self, node: ast.Assign, module: ModuleTable) -> None:
        targets = node.targets
        if (
            ENABLE_IMPLICIT_TYPE_ALIASES
            and len(targets) == 1
            and isinstance(targets[0], ast.Name)
        ):
            # pyre-ignore[16]: `ast.expr` has no attribute `id`
            return self.maybe_set_type_alias(node, targets[0].id)

        for target in targets:
            if isinstance(target, ast.Name):
                self.implicit_decl_names.add(target.id)

    def get_node_data(self, key: AST, data_type: type[TType]) -> TType:
        return cast(TType, self.node_data[key, data_type])

    def get_opt_node_data(self, key: AST, data_type: type[TType]) -> TType | None:
        return cast(Optional[TType], self.node_data.get((key, data_type)))

    def set_node_data(self, key: AST, data_type: type[TType], value: TType) -> None:
        self.node_data[key, data_type] = value

    def mark_known_boolean_test(self, node: ast.expr, *, value: bool) -> None:
        """
        For boolean tests that can be determined during decl-visit, we note the AST nodes
        and the boolean value. This helps us avoid visiting dead code in later passes.
        """
        self.set_node_data(
            node, KnownBoolean, KnownBoolean.TRUE if value else KnownBoolean.FALSE
        )


class IntrinsicModuleTable(ModuleTable):
    """A ModuleTable for modules that are intrinsic in the compiler."""

    def get_child_intrinsic(self, name: str) -> Value | None:
        return self.get_child(name, INTRINSIC_OPT_OUT)


def find_transitive_deps(
    modname: str, all_deps: dict[str, dict[str, set[tuple[str, str]]]]
) -> set[str]:
    """Find all transitive dependency modules of `modname`.

    Given an `alldeps` dictionary of {modname: {name: {(module, name)}}}, return
    the transitive closure of module names depended on by `modname` (not
    including `modname` itself).
    """
    worklist = {dep for deps in all_deps.get(modname, {}).values() for dep in deps}
    ret = set()
    seen = set()
    while worklist:
        dep = worklist.pop()
        seen.add(dep)
        mod, name = dep
        ret.add(mod)
        worklist.update(all_deps.get(mod, {}).get(name, set()).difference(seen))
    ret.discard(modname)
    return ret



# Copyright (c) Meta Platforms, Inc. and affiliates.
# pyre-strict

from __future__ import annotations

import ctypes

from cinderx.static import (
    resolve_primitive_descr,
    TYPED_BOOL,
    TYPED_CHAR,
    TYPED_INT16,
    TYPED_INT32,
    TYPED_INT64,
    TYPED_INT8,
    TYPED_UINT16,
    TYPED_UINT32,
    TYPED_UINT64,
    TYPED_UINT8,
)

_static_to_ctype: dict[int, object] = {
    TYPED_BOOL: ctypes.c_bool,
    TYPED_CHAR: ctypes.c_char,
    # TASK(T130985738): Add support for doubles
    # TYPED_DOUBLE: ctypes.c_double,
    TYPED_INT8: ctypes.c_int8,
    TYPED_INT16: ctypes.c_int16,
    TYPED_INT32: ctypes.c_int32,
    TYPED_INT64: ctypes.c_int64,
    TYPED_UINT8: ctypes.c_uint8,
    TYPED_UINT16: ctypes.c_uint16,
    TYPED_UINT32: ctypes.c_uint32,
    TYPED_UINT64: ctypes.c_uint64,
}


def _create_args(
    signature: tuple[tuple[str, ...], ...], args: tuple[object]
) -> list[object]:
    arg_descrs = signature[:-1]

    n = len(arg_descrs)
    if n != len(args):
        raise RuntimeError(f"invoke_native: {n} args required, got {len(args)}")

    call_args = []
    for i in range(n):
        descr = arg_descrs[i]
        arg = args[i]
        primitive_type = resolve_primitive_descr(descr)
        try:
            ctypes_type = _static_to_ctype[primitive_type]
        except KeyError as e:
            raise RuntimeError(f"Unsupported primitive type: {descr}") from e
        else:
            # pyre-ignore[29]: object is not callable
            call_args.append(ctypes_type(arg))
    return call_args


def invoke_native(
    libname: str,
    symbol: str,
    signature: tuple[tuple[str, ...], ...],
    args: tuple[object],
) -> int:
    # This is basically just a `dlopen()` under the hood
    lib = ctypes.CDLL(libname)

    # Python wrapper over `dlsym`
    fn = getattr(lib, symbol)

    call_args = _create_args(signature, args)

    res = fn(*call_args)
    return res



# Copyright (c) Meta Platforms, Inc. and affiliates.
# pyre-strict

TYPED_INT8 = 0
TYPED_INT16 = 0
TYPED_INT32 = 0
TYPED_INT64 = 0
TYPED_UINT8 = 0
TYPED_UINT16 = 0
TYPED_UINT32 = 0
TYPED_UINT64 = 0
TYPED_INT_64BIT = 0
TYPED_DOUBLE = 0
TYPED_SINGLE = 0
TYPED_BOOL = 0
TYPED_CHAR = 0
TYPED_OBJECT = 0

try:
    from cinderx.static import (  # noqa: F401
        set_type_code,
        TYPED_BOOL,
        TYPED_CHAR,
        TYPED_DOUBLE,
        TYPED_INT16,
        TYPED_INT32,
        TYPED_INT64,
        TYPED_INT8,
        TYPED_INT_64BIT,
        TYPED_OBJECT,
        TYPED_SINGLE,
        TYPED_UINT16,
        TYPED_UINT32,
        TYPED_UINT64,
        TYPED_UINT8,
    )
except ImportError:

    def set_type_code(func: object, code: object) -> None:
        pass


def type_code(code: int) -> object:
    def inner(func: object) -> object:
        set_type_code(func, code)
        return func

    return inner



# Copyright (c) Meta Platforms, Inc. and affiliates.
# pyre-strict

import sys
from importlib.abc import Loader
from importlib.machinery import (
    BYTECODE_SUFFIXES,
    EXTENSION_SUFFIXES,
    ExtensionFileLoader,
    FileFinder,
    SOURCE_SUFFIXES,
    SourcelessFileLoader,
)
from typing import Callable, Iterable, Mapping

from cinderx.compiler.static.pyrefly_compiler import PyreflyCompiler
from cinderx.compiler.static.pyrefly_info import Pyrefly

from .compiler import Compiler, TIMING_LOGGER_TYPE
from .loader import StrictSourceFileLoader


OPT_OUT_LIST: set[str] | None = None
OPT_IN_LIST: set[str] | None = None


class PyreflyLoader(StrictSourceFileLoader):
    pyrefly_type_dir: str | None = None

    @classmethod
    def ensure_compiler(
        cls,
        path: Iterable[str],
        stub_path: str,
        allow_list_prefix: Iterable[str],
        allow_list_exact: Iterable[str],
        log_time_func: Callable[[], TIMING_LOGGER_TYPE] | None,
        enable_patching: bool = False,
        allow_list_regex: Iterable[str] | None = None,
    ) -> Compiler:
        if (comp := cls.compiler) is None:
            if cls.pyrefly_type_dir is not None:
                pyrefly = Pyrefly(cls.pyrefly_type_dir)
            else:
                pyrefly = None
            comp = cls.compiler = PyreflyCompiler(
                static_opt_out=OPT_OUT_LIST,
                static_opt_in=OPT_IN_LIST,
                path=path,
                stub_path=stub_path,
                allow_list_prefix=allow_list_prefix,
                allow_list_exact=allow_list_exact,
                pyrefly=pyrefly,
                log_time_func=log_time_func,
                enable_patching=enable_patching,
                allow_list_regex=allow_list_regex or [],
            )
        return comp

    def should_force_strict(self) -> bool:
        if OPT_IN_LIST is not None:
            return self.name in OPT_IN_LIST
        if OPT_OUT_LIST is not None:
            return self.name not in OPT_OUT_LIST
        return True


class PyreflyLoaderWithPatching(PyreflyLoader):
    def __init__(
        self,
        fullname: str,
        path: str,
        import_path: Iterable[str] | None = None,
        stub_path: str | None = None,
        allow_list_prefix: Iterable[str] | None = None,
        allow_list_exact: Iterable[str] | None = None,
        enable_patching: bool = True,
        log_source_load: Callable[[str, str | None, bool], None] | None = None,
        init_cached_properties: None
        | (
            Callable[
                [Mapping[str, str | tuple[str, bool]]],
                Callable[[type[object]], type[object]],
            ]
        ) = None,
        log_time_func: Callable[[], TIMING_LOGGER_TYPE] | None = None,
        use_py_compiler: bool = False,
        # The regexes are parsed on the C++ side, so re.Pattern is not accepted.
        allow_list_regex: Iterable[str] | None = None,
    ) -> None:
        super().__init__(
            fullname,
            path,
            import_path,
            stub_path,
            allow_list_prefix,
            allow_list_exact,
            enable_patching,
            log_source_load,
            init_cached_properties,
            log_time_func,
            use_py_compiler,
            allow_list_regex,
        )


def _get_supported_file_loaders(
    enable_patching: bool = False,
) -> list[tuple[type[Loader], list[str]]]:
    """Returns a list of file-based module loaders.

    Each item is a tuple (loader, suffixes).
    """
    extensions = ExtensionFileLoader, EXTENSION_SUFFIXES
    source = (
        (PyreflyLoaderWithPatching if enable_patching else PyreflyLoader),
        SOURCE_SUFFIXES,
    )
    bytecode = SourcelessFileLoader, BYTECODE_SUFFIXES
    return [extensions, source, bytecode]


def install(
    enable_patching: bool = False,
    pyrefly_type_dir: str | None = None,
    opt_in_list: set[str] | None = None,
    opt_out_list: set[str] | None = None,
) -> None:
    """Installs a loader which is capable of loading and validating strict modules"""
    PyreflyLoader.pyrefly_type_dir = pyrefly_type_dir
    supported_loaders = _get_supported_file_loaders(enable_patching)

    global OPT_IN_LIST, OPT_OUT_LIST
    OPT_IN_LIST = opt_in_list
    OPT_OUT_LIST = opt_out_list

    for index, hook in enumerate(sys.path_hooks):
        if not isinstance(hook, type):
            sys.path_hooks.insert(index, FileFinder.path_hook(*supported_loaders))
            break
    else:
        sys.path_hooks.insert(0, FileFinder.path_hook(*supported_loaders))

    # We need to clear the path_importer_cache so that our new FileFinder will
    # start being used for existing directories we've loaded modules from.
    sys.path_importer_cache.clear()



# Copyright (c) Meta Platforms, Inc. and affiliates.

# pyre-strict

from __future__ import annotations

import ast
import logging
import os
import symtable
import sys
from contextlib import nullcontext
from dataclasses import dataclass
from symtable import SymbolTable as PythonSymbolTable
from types import CodeType
from typing import Callable, ContextManager, final, Iterable

from ..errors import TypedSyntaxError
from ..pycodegen import compile_code
from ..static import StaticCodeGenerator
from ..static.compiler import Compiler as StaticCompiler
from ..static.module_table import ModuleTable
from . import strict_compile
from .common import StrictModuleError
from .flag_extractor import FlagExtractor, Flags
from .rewriter import remove_annotations, rewrite


TIMING_LOGGER_TYPE = Callable[[str, str, str], ContextManager[None]]


@dataclass(frozen=True)
class SourceInfo:
    """Data about a Python module source file."""

    # module name
    name: str
    # import path prefix within which the module was found
    prefix: str
    # path to module file, relative to import path prefix
    relative_path: str
    # should always be os.path.join(prefix, relative_path)
    path: str
    # data we might need for pyc invalidation
    source: bytes | None
    mtime: int
    size: int


@dataclass(frozen=True)
class Dependencies:
    static: list[SourceInfo]
    nonstatic: list[str]


NOT_FOUND = object()


@final
class Compiler(StaticCompiler):
    def __init__(
        self,
        import_path: Iterable[str],
        stub_root: str,
        allow_list_prefix: Iterable[str],
        allow_list_exact: Iterable[str],
        log_time_func: Callable[[], TIMING_LOGGER_TYPE] | None = None,
        raise_on_error: bool = False,
        enable_patching: bool = False,
        use_py_compiler: bool = False,
        allow_list_regex: Iterable[str] | None = None,
    ) -> None:
        super().__init__(StaticCodeGenerator)
        self.import_path: list[str] = list(import_path)
        self.stub_root = stub_root
        self.allow_list_prefix = allow_list_prefix
        self.allow_list_exact = allow_list_exact
        self.allow_list_regex: Iterable[str] = allow_list_regex or []
        self.verbose = bool(
            os.getenv("PYTHONSTRICTVERBOSE")
            or sys._xoptions.get("strict-verbose") is True
        )
        self.disable_analysis = bool(
            os.getenv("PYTHONSTRICTDISABLEANALYSIS")
            or sys._xoptions.get("strict-disable-analysis") is True
        )
        self.raise_on_error = raise_on_error
        self.log_time_func = log_time_func
        self.enable_patching = enable_patching
        self.not_static: set[str] = set()
        self.use_py_compiler = use_py_compiler
        self.original_builtins: dict[str, object] = dict(__builtins__)
        self.logger: logging.Logger = self._setup_logging()
        self.sources: dict[str, SourceInfo | None] = {}

    def get_dependencies(self, name: str) -> Dependencies:
        """Get static and non-static dependencies of given module."""
        mod = self.modules.get(name)
        if mod is None:
            return Dependencies(static=[], nonstatic=[])

        static = []
        nonstatic = []
        for depmod in mod.get_dependencies():
            if isinstance(depmod, ModuleTable):
                # changes to these should be reflected in our compiler magic
                # number, and don't need to be recorded as dependencies
                if depmod.name in self.intrinsic_modules:
                    continue

                source_info = self.sources.get(depmod.name)
                if source_info is None:
                    # the dependency may have been previously compiled as the
                    # main module in the same Compiler, in which case we'll have
                    # a ModuleTable already but no SourceInfo
                    source_info = self.get_source(depmod.name)
                assert source_info is not None
                static.append(source_info)
            else:  # UnknownModule, non-static
                nonstatic.append(depmod.name)

        return Dependencies(static=static, nonstatic=nonstatic)

    def import_module(self, name: str, optimize: int) -> ModuleTable | None:
        res = self.modules.get(name)
        if res is not None:
            return res

        if name in self.not_static:
            return None

        source_info = self.get_source(name)
        if source_info is None:
            return None
        source = source_info.source
        assert source is not None
        filename = source_info.relative_path

        pyast = ast.parse(source)
        flags = FlagExtractor().get_flags(pyast)

        if name not in self.modules:
            if flags.is_static:
                # pyre-fixme[6]: For 1st argument expected `str` but got
                #  `Union[bytes, str]`.
                symbols = symtable.symtable(source, filename, "exec")
                root = pyast

                if filename.endswith(".pyi"):
                    root = remove_annotations(root)

                root = self._get_rewritten_ast(name, root, symbols, filename, optimize)
                log = self.log_time_func
                ctx = (
                    log()(name, filename, "declaration_visit") if log else nullcontext()
                )
                with ctx:
                    root = self.add_module(name, filename, root, source, optimize)
            else:
                self.not_static.add(name)

        return self.modules.get(name)

    def _get_rewritten_ast(
        self,
        name: str,
        root: ast.Module,
        symbols: PythonSymbolTable,
        filename: str,
        optimize: int,
    ) -> ast.Module:
        return rewrite(
            root,
            symbols,
            filename,
            name,
            optimize=optimize,
            is_static=True,
            builtins=self.original_builtins,
        )

    def _setup_logging(self) -> logging.Logger:
        logger = logging.Logger(__name__)
        if self.verbose:
            logger.setLevel(logging.DEBUG)
        return logger

    def get_flags(
        self, module_name: str, pyast: ast.Module, override_flags: Flags
    ) -> Flags:
        return FlagExtractor().get_flags(pyast).merge(override_flags)

    def load_compiled_module_from_source(
        self,
        source: str | bytes,
        filename: str,
        name: str,
        optimize: int,
        submodule_search_locations: list[str] | None = None,
        override_flags: Flags | None = None,
    ) -> tuple[CodeType | None, bool, bool]:
        pyast = ast.parse(source)
        # pyre-fixme[6]: For 1st argument expected `str` but got `Union[bytes, str]`.
        symbols = symtable.symtable(source, filename, "exec")
        # pyre-fixme[6]: For 1st argument expected `Flag` but got `Optional[Flags]`.
        flags = self.get_flags(name, pyast, override_flags)

        if not flags.is_static and not flags.is_strict:
            code = self._compile_basic(name, pyast, filename, optimize)
            return (code, False, False)

        if flags.is_static:
            code = self._compile_static(
                pyast, source, symbols, filename, name, optimize
            )
            return (code, flags.is_strict, True)
        else:
            code = self._compile_strict(
                pyast, source, symbols, filename, name, optimize
            )
            return (code, flags.is_strict, False)

    def get_source(self, name: str, *, need_contents: bool = True) -> SourceInfo | None:
        source_info = self.sources.get(name, NOT_FOUND)
        if source_info is NOT_FOUND:
            self.sources[name] = source_info = self._get_source(
                name, need_contents=need_contents
            )
        assert source_info is None or isinstance(source_info, SourceInfo)
        if source_info and source_info.source is None and need_contents:
            # We don't need to re-do the sys.path finding, but we should update
            # the size and mtime to match the current source.
            self.sources[name] = source_info = self._get_source_info(
                name, source_info.prefix, source_info.relative_path, need_contents=True
            )
        return source_info

    def _get_source(self, name: str, *, need_contents: bool) -> SourceInfo | None:
        module_path = name.replace(".", os.sep)

        for path in self.import_path:
            source_info = self._get_source_info(
                name,
                path,
                module_path + os.sep + "__init__.py",
                need_contents=need_contents,
            )
            if source_info is not None:
                return source_info

            source_info = self._get_source_info(
                name, path, module_path + ".py", need_contents=need_contents
            )
            if source_info is not None:
                return source_info

        for path in self.import_path:
            source_info = self._get_source_info(
                name, path, module_path + ".pyi", need_contents=need_contents
            )
            if source_info is not None:
                return source_info

        return None

    def _get_source_info(
        self, name: str, prefix: str, relpath: str, *, need_contents: bool
    ) -> SourceInfo | None:
        path = os.path.join(prefix, relpath)
        if not os.path.exists(path):
            return None
        st = os.stat(path)
        source = None
        if need_contents:
            with open(path, "rb") as f:
                source = f.read()
        return SourceInfo(
            name=name,
            prefix=prefix,
            relative_path=relpath,
            path=path,
            source=source,
            mtime=int(st.st_mtime),
            size=st.st_size,
        )

    def _compile_basic(
        self, name: str, root: ast.Module, filename: str, optimize: int
    ) -> CodeType:
        compile_method = compile_code if self.use_py_compiler else compile
        result = compile_method(
            root,
            filename,
            "exec",
            optimize=optimize,
        )

        assert isinstance(result, CodeType)
        return result

    def _compile_strict(
        self,
        root: ast.Module,
        source: str | bytes,
        symbols: PythonSymbolTable,
        filename: str,
        name: str,
        optimize: int,
    ) -> CodeType:
        tree = rewrite(
            root,
            symbols,
            filename,
            name,
            optimize=optimize,
            builtins=self.original_builtins,
        )
        return strict_compile(
            name, filename, tree, source, optimize, self.original_builtins
        )

    def _compile_static(
        self,
        root: ast.Module,
        source: str | bytes,
        symbols: PythonSymbolTable,
        filename: str,
        name: str,
        optimize: int,
    ) -> CodeType | None:
        root = self.ast_cache.get(source) or self._get_rewritten_ast(
            name, root, symbols, filename, optimize
        )
        try:
            log = self.log_time_func
            ctx = log()(name, filename, "compile") if log else nullcontext()
            with ctx:
                return self.compile(
                    name,
                    filename,
                    root,
                    source,
                    optimize,
                    enable_patching=self.enable_patching,
                    builtins=self.original_builtins,
                )
        except TypedSyntaxError as e:
            err = StrictModuleError(
                e.msg or "unknown error during static compilation",
                e.filename or filename,
                e.lineno or 1,
                0,
            )

            if self.raise_on_error:
                raise err

            return None
